#!/usr/bin/env python
#
# fsleyes.py - Image viewer.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the entry point to *FSLeyes*, the FSL image viewer.

Call the :func:`main` function to start the main FSLeyes application.

The :func:`embed` function can be called to open a :class:`.FSLeyesFrame`
within an existing application.

See the :mod:`fsleyes` package documentation for more details on ``fsleyes``.


.. note:: Even though ``fsleyes`` (this module) and :mod:`fsleyes.render` (the
          off-screen renderer) are intended to be separate applications, the
          current version of PyInstaller (3.x) does not support bundling of
          multiple executables
          (https://github.com/pyinstaller/pyinstaller/issues/1527).

          So at this point in time, :mod:`.fsleyes.render` can be invoked via
          ``fsleyes.main`` by passing ``'render'`` as the first argument,
          e.g.::

              python -m fsleyes.main render ...
"""


import functools as ft
import os.path   as op
import              os
import              sys
import              signal
import              logging
import              textwrap

import wx
import wx.adv

from   fsl.utils.platform import platform as fslplatform
import fsl.utils.idle                     as idle
import fsleyes_widgets                    as fwidgets
import fsleyes_widgets.utils.status       as status

import                       fsleyes
import fsleyes.strings    as strings
import fsleyes.splash     as fslsplash
import fsleyes.cliserver  as cliserver
import fsleyes.colourmaps as colourmaps


# wx.ModalDialogHook does not exist in wxPython < 4

if fwidgets.wxFlavour() in (fwidgets.WX_PYTHON, fwidgets.WX_UNKNOWN):
    class ModalDialogHook(object):
        def Register(self):
            pass

    wx.ModalDialogHook = ModalDialogHook


log = logging.getLogger(__name__)


class FSLeyesApp(wx.App):
    """FSLeyes-specific sub-class of ``wx.App``. """


    class ModalHook(wx.ModalDialogHook):
        """Keeps track of any modal dialogs/windows that are opened.

        Modal dialogs can interfere with shutdown, as they run their own event
        loop. Therefore we keep a reference is kept to all opened modal
        dialogs, so we can manually shut them down if needed (see the
        :func:`main` function).
        """

        def __init__(self, *args, **kwargs):
            wx.ModalDialogHook.__init__(self, *args, **kwargs)
            self.modals = set()

        def Enter(self, dlg):
            self.modals.add(dlg)
            return wx.ID_NONE

        def Exit(self, dlg):
            self.modals.discard(dlg)


    def __init__(self):
        """Create a ``FSLeyesApp``. """

        self.__overlayList = None
        self.__displayCtx  = None

        # On macOS, when the user drags a file onto the FSLeyes window,
        # or onto the FSLeyes.app icon, the file path will be passed to
        # MacOpenFiles method. But that method may be called very early
        # on in the startup process, before the DisplayContext and
        # OverlayList have been created. So when this happens, we cache
        # the files here, and then open them when the
        # SetOverlayListAndDisplayContext method gets called.
        self.__filesToOpen = []

        self.__modalHook = FSLeyesApp.ModalHook()
        self.__modalHook.Register()

        wx.App.__init__(self, clearSigInt=False)

        self.SetAppName('FSLeyes')

        try:
            self.__icon = wx.adv.TaskBarIcon(iconType=wx.adv.TBI_DOCK)
            self.__icon.SetIcon(wx.Icon(
                op.join(fsleyes.assetDir, 'assets', 'icons', 'app_icon.png')))
        except Exception:
            self.__icon = None


    @property
    def modals(self):
        """Returns a list of all currently open modal windows. """
        return list(self.__modalHook.modals)


    def SetOverlayListAndDisplayContext(self, overlayList, displayCtx):
        """References to the :class:`.OverlayList` and master
        :class:`.DisplayContext` must be passed to the ``FSLeyesApp`` via this
        method.
        """
        self.__overlayList = overlayList
        self.__displayCtx  = displayCtx

        # MacOpenFiles was called before the
        # overlaylist/dc were created, and
        # queued some files that need to be
        # opened
        if len(self.__filesToOpen) > 0:
            wx.CallAfter(self.MacOpenFiles, self.__filesToOpen)
            self.__filesToOpen = None


    def MacReopenApp(self):
        """On OSX, make sure that the FSLeyes frame is restored if it is
        minimised, and (e.g.) the dock icon is clicked.
        """

        frame = self.GetTopWindow()
        frame.Iconize(False)
        frame.Raise()


    def MacOpenFile(self, filename):
        """On OSX, support opening files via context menu, and files dropped
        on the application icon.
        """
        self.MacOpenFiles([filename])


    def MacOpenURL(self, url):
        """On OSX, support opening files via a ``fsleyes://`` url. """

        if self.__overlayList is None:
            return

        import fsleyes_widgets.utils.status     as status
        import fsleyes.strings                  as strings
        import fsleyes.parseargs                as parseargs
        import fsleyes.actions.applycommandline as applycommandline

        errTitle = strings.titles[  self, 'openURLError']
        errMsg   = strings.messages[self, 'openURLError']

        with status.reportIfError(errTitle, errMsg):
            applycommandline.applyCommandLineArgs(
                self.__overlayList,
                self.__displayCtx,
                parseargs.fsleyesUrlToArgs(url))


    def MacOpenFiles(self, filenames):
        """On OSX, support opening files via context menu, and files dropped
        on the application icon.
        """

        # OverlayList has not yet been created -
        # queue the files to open them later
        # in SetOverlayListAndDisplayContext
        if self.__overlayList is None:
            self.__filesToOpen.extend(filenames)
            return

        import fsleyes.actions.loadoverlay as loadoverlay
        import fsleyes.autodisplay         as autodisplay

        def onLoad(paths, overlays):

            if len(overlays) == 0:
                return

            self.__overlayList.extend(overlays)

            if self.__displayCtx.autoDisplay:
                for overlay in overlays:
                    autodisplay.autoDisplay(overlay,
                                            self.__overlayList,
                                            self.__displayCtx)

        loadoverlay.loadOverlays(
            filenames,
            onLoad=onLoad,
            inmem=self.__displayCtx.loadInMemory)


def main(args=None):
    """*FSLeyes* entry point. Shows a :class:`.FSLeyesSplash` screen, parses
    command line arguments, and shows a :class:`.FSLeyesFrame`. Returns
    an exit code.
    """

    if args is None:
        args = sys.argv[1:]

    # Hack to allow render to
    # be called via fsleyes.main
    if len(args) >= 1 and args[0] == 'render':
        import fsleyes.render as render
        render.main(args[1:])
        sys.exit(0)

    # the fsleyes.initialise function figures
    # out the path to asset files (e.g. cmaps)
    fsleyes.initialise()

    # Hook which allows us to run a jupyter
    # notebook server from an existing FSLeyes
    # instance
    if len(args) >= 1 and args[0] == 'notebook':
        from fsleyes.actions.notebook import nbmain
        fsleyes.configLogging()
        sys.exit(nbmain(args))

    # initialise colour maps - this must be
    # done before parsing arguments, as if
    # the user asks for help, available
    # colourmaps/luts will be listed.
    colourmaps.init()

    # Function to bootstrap the GUI - keep
    # reading below.
    def initgui():

        # First thing's first. Create a wx.App,
        # and initialise the FSLeyes package.
        app = FSLeyesApp()

        # Create a splash screen frame
        splash = fslsplash.FSLeyesSplash(None)
        return app, splash

    # If it looks like the user is asking for
    # help, or using cliserver to pass arguments
    # to an existing FSLeyes instance, then we
    # parse command line arguments before
    # creating a wx.App and showing the splash
    # screen. This means that FSLeyes help/
    # version information can be retrieved
    # without a display, and hopefully fairly
    # quickly.
    #
    # Otherwise we create the app and splash
    # screen first, so the splash screen gets
    # shown as soon as possible. Arguments
    # will get parsed in the init function below.
    #
    # The argparse.Namespace object is kept in a
    # list so it can be shared between the sub-
    # functions below
    #
    # If argument parsing bombs out, we put the
    # exit code here and return it at the bottom.
    namespace = [None]
    exitCode  = [0]

    # user asking for help - parse args first
    if (len(args) > 0) and (args[0] in ('-V',
                                        '-h',
                                        '-fh',
                                        '-cs',
                                        '--version',
                                        '--help',
                                        '--fullhelp',
                                        '--cliserver')):
        namespace   = [parseArgs(args)]
        app, splash = initgui()

    # otherwise parse arguments on wx.MainLoop
    # below
    else:
        app, splash = initgui()

    # We are going do all processing on the
    # wx.MainLoop, so the GUI can be shown
    # as soon as possible, and because it is
    # difficult to force immediate GUI
    # refreshes when not running on the main
    # loop - this is important for FSLeyes,
    # which displays status updates to the
    # user while it is loading overlays and
    # setting up the interface.
    #
    # All of the work is defined in a series
    # of functions, which are chained together
    # via ugly callbacks, but which are
    # ultimately scheduled and executed on the
    # wx main loop.
    def init(splash):

        # See FSLeyesSplash.Show
        # for horribleness.
        splash.Show()

        # Parse command line arguments if necessary.
        # If arguments are invalid, the parseargs
        # module will raise SystemExit.
        try:
            if namespace[0] is None:

                errmsg   = strings.messages['main.parseArgs.error']
                errtitle = strings.titles[  'main.parseArgs.error']
                with status.reportIfError(errtitle, errmsg, raiseError=True):
                    namespace[0] = parseArgs(args)

        # But the wx.App.MainLoop eats SystemExit
        # exceptions for unknown reasons, and
        # causes the application to exit
        # immediately. This makes testing FSLeyes
        # (e.g. code coverage) impossible. So I'm
        # catching SystemExit here, and then
        # telling the wx.App to exit gracefully.
        except (SystemExit, Exception) as e:
            app.ExitMainLoop()
            exitCode[0] = getattr(e, 'code', 1)
            return

        # Configure logging (this has to be done
        # after cli arguments have been parsed,
        # but before initialise is called).
        fsleyes.configLogging(namespace[0].verbose, namespace[0].noisy)

        # Initialise sub-modules/packages. The
        # buildGui function is passed through
        # as a callback, which gets called when
        # initialisation is complete.
        initialise(splash, namespace[0], buildGui)

    def buildGui():

        # Now the main stuff - create the overlay
        # list and the master display context,
        # and then create the FSLeyesFrame.
        overlayList, displayCtx = makeDisplayContext(namespace[0], splash)
        app.SetOverlayListAndDisplayContext(overlayList, displayCtx)
        frame = makeFrame(namespace[0],
                          displayCtx,
                          overlayList,
                          splash,
                          [shutdown])

        app.SetTopWindow(frame)
        frame.Show()

        # Check that $FSLDIR is set, complain
        # to the user if it isn't
        if not namespace[0].skipfslcheck:
            wx.CallAfter(fslDirWarning, frame)

        # Check for updates. Ignore point
        # releases, otherwise users might
        # get swamped with update notifications.
        if namespace[0].updatecheck:
            import fsleyes.actions.updatecheck as updatecheck
            wx.CallAfter(updatecheck.UpdateCheckAction(),
                         showUpToDateMessage=False,
                         showErrorMessage=False,
                         ignorePoint=False)

        # start notebook server
        if namespace[0].notebookFile is not None:
            namespace[0].notebook     = True
            namespace[0].notebookFile = op.abspath(namespace[0].notebookFile)
        if namespace[0].notebook:
            from fsleyes.actions.notebook import NotebookAction
            frame.menuActions[NotebookAction](namespace[0].notebookFile)

        # start CLI server
        if namespace[0].cliserver:
            cliserver.runserver(overlayList, displayCtx)

    # Shut down cleanly on sigint/sigterm.
    # We do this so that any functions
    # registered with atexit will actually
    # get called.
    nsignals = [0]

    def sigHandler(signo, frame):
        log.debug('Signal received - FSLeyes is shutting down...')

        # first signal - try to exit cleanly
        if nsignals[0] == 0:
            nsignals[0] += 1
            exitCode[0]  = signo

            # kill any modal windows
            # that are open
            for mdlg in app.modals:
                mdlg.EndModal(wx.ID_CANCEL)

            wx.CallAfter(app.ExitMainLoop)

        # subsequent signals - exit immediately
        else:
            sys.exit(signo)

    signal.signal(signal.SIGINT,  sigHandler)
    signal.signal(signal.SIGTERM, sigHandler)

    # Note: If no wx.Frame is created, the
    # wx.MainLoop call will exit immediately,
    # even if we have scheduled something via
    # wx.CallAfter. In this case, we have
    # already created the splash screen, so
    # all is well.
    wx.CallAfter(init, splash)

    # under mac, use appnope to make sure
    # we don't get put to sleep. This is
    # primarily for the jupyter notebook
    # integration - if the user is working
    # with a notebook in the web browser,
    # macos might put FSLeyes to sleep,
    # causing the kernel to become
    # unresponsive.
    try:
        import appnope
        appnope.nope()
    except ImportError:
        pass
    app.MainLoop()
    return exitCode[0]


def embed(mkFrame=True, **kwargs):
    """Initialise FSLeyes and create a :class:`.FSLeyesFrame`, when
    running within another application.

    .. note:: In most cases, this function must be called from the
              ``wx.MainLoop``.

    :arg mkFrame: Defaults to ``True``. If ``False``, FSLeyes is
                  initialised, but a :class:`.FSLeyesFrame` is not created.
                  If you set this to ``False``, you must ensure that a
                  ``wx.App`` object exists before calling this function.

    :returns:     A tuple containing:

                   - The :class:`.OverlayList`
                   - The master :class:`.DisplayContext`
                   - The :class:`.FSLeyesFrame` (or ``None``, if
                     ``makeFrame is False``).

    All other arguments are passed to :meth:`.FSLeyesFrame.__init__`.
    """

    import fsleyes_props          as props
    import fsleyes.gl             as fslgl
    import fsleyes.frame          as fslframe
    import fsleyes.overlay        as fsloverlay
    import fsleyes.displaycontext as fsldc

    # initialise must be called before
    # a FSLeyesApp gets created, as it
    # tries to access app_icon.png
    fsleyes.initialise()

    app    = wx.GetApp()
    ownapp = app is None

    if ownapp and (mkFrame is False):
        raise RuntimeError('If mkFrame is False, you '
                           'must create a wx.App before '
                           'calling fsleyes.main.embed')
    if ownapp:
        app = FSLeyesApp()

    colourmaps.init()
    props.initGUI()

    called = [False]
    ret    = [None]

    def until():
        return called[0]

    def ready():
        fslgl.bootstrap()

        overlayList = fsloverlay.OverlayList()
        displayCtx  = fsldc.DisplayContext(overlayList)

        if mkFrame:
            frame = fslframe.FSLeyesFrame(
                None, overlayList, displayCtx, **kwargs)
        else:
            frame = None

        if ownapp:
            app.SetOverlayListAndDisplayContext(overlayList, displayCtx)
            # Keep a ref to prevent the app from being GC'd
            frame._embed_app = app

        called[0] = True
        ret[0]    = (overlayList, displayCtx, frame)

    fslgl.getGLContext(ready=ready, raiseErrors=True)
    idle.block(10, until=until)

    if ret[0] is None:
        raise RuntimeError('Failed to start FSLeyes')
    return ret[0]


def initialise(splash, namespace, callback):
    """Called by :func:`main`. Bootstraps/Initialises various parts of
    *FSLeyes*.

    The ``callback`` function is asynchronously called when the initialisation
    is complete.

    :arg splash:    The :class:`.FSLeyesSplash` screen.

    :arg namespace: The ``argparse.Namespace`` object containing parsed
                    command line arguments.

    :arg callback:  Function which is called when initialisation is done.
    """

    import fsl.utils.settings as fslsettings
    import fsleyes_props      as props
    import fsleyes.gl         as fslgl

    props.initGUI()

    # The save/load directory defaults
    # to the current working directory.
    curDir = op.normpath(os.getcwd())

    fslsettings.write('loadSaveOverlayDir', curDir)

    # Initialise silly things
    if namespace.bumMode:
        import fsleyes.icons as icons
        icons.BUM_MODE = True

    # Set notebook server port
    fslsettings.write('fsleyes.notebook.port', namespace.notebookPort)

    # This is called by fsleyes.gl.getGLContext
    # when the GL context is ready to be used.
    def realCallback():
        fslgl.bootstrap(namespace.glversion)
        callback()

    try:
        # Force the creation of a wx.glcanvas.GLContext object,
        # and initialise OpenGL version-specific module loads.
        fslgl.getGLContext(ready=realCallback)

    except Exception:
        log.error('Unable to initialise OpenGL!', exc_info=True)
        splash.Destroy()
        sys.exit(1)


def shutdown(ev=None):
    """Called when FSLeyes exits normally (i.e. the user closes the window).
    Does some final clean-up before exiting.

    This function is used as a wxpython event handler, so it accepts an ``ev``
    arguments, but ignores its value.
    """

    if ev is not None:
        ev.Skip()

    import fsl.utils.settings as fslsettings
    import fsleyes.gl         as fslgl

    # Clear the cached directory for loading/saving
    # files - when FSLeyes starts up, we want it to
    # default to the current directory.
    fslsettings.delete('loadSaveOverlayDir')

    # Shut down the GL rendering context
    fslgl.shutdown()


def parseArgs(argv):
    """Parses the given ``fsleyes`` command line arguments. See the
    :mod:`.parseargs` module for details on the ``fsleyes`` command
    line interface.

    :arg argv: command line arguments for ``fsleyes``.
    """

    import fsleyes.parseargs as parseargs
    import fsleyes.layouts   as layouts
    import fsleyes.version   as version

    parser = parseargs.ArgumentParser(
        add_help=False,
        formatter_class=parseargs.FSLeyesHelpFormatter)

    serveraction = ft.partial(cliserver.CLIServerAction, allArgs=argv)

    parser.add_argument('-r', '--runscript',
                        metavar='SCRIPTFILE',
                        help='Run custom FSLeyes script')
    parser.add_argument('-cs', '--cliserver',
                        action=serveraction,
                        help='Pass all command-line arguments '
                             'to a single FSLeyes instance')

    # We include the list of available
    # layouts in the help description
    allLayouts  = list(layouts.BUILT_IN_LAYOUTS.keys()) + \
                  list(layouts.getAllLayouts())
    name        = 'fsleyes'
    prolog      = 'FSLeyes version {}\n'.format(version.__version__)
    description = textwrap.dedent("""\
        FSLeyes - the FSL image viewer.

        Use the '--scene' option to load a saved layout ({layouts}).

        If no '--scene' is specified, a default layout is shown or the
        previous layout is restored. If a script is provided via
        the '--runscript' argument, it is assumed that the script sets
        up the scene.
        """.format(layouts=', '.join(allLayouts)))

    # Options for configuring the scene are
    # managed by the parseargs module
    return parseargs.parseArgs(parser,
                               argv,
                               name,
                               prolog=prolog,
                               desc=description,
                               argOpts=['-r', '--runscript'])


def makeDisplayContext(namespace, splash):
    """Creates the top-level *FSLeyes* :class:`.DisplayContext` and
    :class:`.OverlayList` .

    This function does the following:

     1. Creates the :class:`.OverlayList` and the top level
        :class:`.DisplayContext`.

     2. Loads and configures all of the overlays which were passed in on the
        command line.

    :arg namesace: Parsed command line arguments (see :func:`parseArgs`).

    :arg splash:   The :class:`.FSLeyesSplash` frame, created in :func:`init`.

    :returns: a tuple containing:
                - the :class:`.OverlayList`
                - the master :class:`.DisplayContext`
    """

    import fsleyes_widgets.utils.status as status
    import fsleyes.overlay              as fsloverlay
    import fsleyes.parseargs            as parseargs
    import fsleyes.displaycontext       as displaycontext

    # Splash status update must be
    # performed on the main thread.
    def splashStatus(msg):
        wx.CallAfter(splash.SetStatus, msg)

    # Redirect status updates
    # to the splash frame
    status.setTarget(splashStatus)

    # Create the overlay list (only one of these
    # ever exists) and the master DisplayContext.
    # A new DisplayContext instance will be
    # created for every new view that is opened
    # in the FSLeyesFrame, but all child
    # DisplayContext instances will be linked to
    # this master one.
    overlayList = fsloverlay.OverlayList()
    displayCtx  = displaycontext.DisplayContext(overlayList)

    log.debug('Created overlay list and master DisplayContext ({})'.format(
        id(displayCtx)))

    # Load the images - the splash screen status will
    # be updated with the currently loading overlay name.
    parseargs.applyMainArgs(   namespace, overlayList, displayCtx)
    parseargs.applyOverlayArgs(namespace, overlayList, displayCtx)

    return overlayList, displayCtx


def makeFrame(namespace, displayCtx, overlayList, splash, closeHandlers):
    """Creates the *FSLeyes* interface.

    This function does the following:

     1. Creates the :class:`.FSLeyesFrame` the top-level frame for ``fsleyes``.

     2. Configures the frame according to the command line arguments (e.g.
        ortho or lightbox view).

     3. Destroys the splash screen that was created by the :func:`context`
        function.

    :arg namespace:     Parsed command line arguments, as returned by
                        :func:`parseArgs`.

    :arg displayCtx:    The  :class:`.DisplayContext`, as created and returned
                        by :func:`makeDisplayContext`.

    :arg overlayList:   The :class:`.OverlayList`, as created and returned by
                        :func:`makeDisplayContext`.

    :arg splash:        The :class:`.FSLeyesSplash` frame.

    :arg closeHandlers: List of event handlers to be called when the
                        ``FSLeyesFrame`` closes.

    :returns: the :class:`.FSLeyesFrame` that was created.
    """

    import fsl.utils.idle               as idle
    import fsleyes_widgets.utils.status as status
    import fsleyes.parseargs            as parseargs
    import fsleyes.frame                as fsleyesframe
    import fsleyes.displaycontext       as fsldisplay
    import fsleyes.layouts              as layouts
    import fsleyes.views.canvaspanel    as canvaspanel

    # Set up the frame scene (a.k.a. layout)
    # The scene argument can be:
    #
    #   - The name of a saved (or built-in) layout
    #
    #   - None, in which case the default or previous
    #     layout is restored, unless a custom script
    #     has been provided.
    script = namespace.runscript
    scene  = namespace.scene

    # If a scene/layout or custom script
    # has not been specified, the default
    # behaviour is to restore the previous
    # frame layout.
    restore = (scene is None) and (script is None)

    status.update('Creating FSLeyes interface...')

    frame = fsleyesframe.FSLeyesFrame(
        None,
        overlayList,
        displayCtx,
        restore,
        True,
        fontSize=namespace.fontSize,
        closeHandlers=closeHandlers)

    # Allow files to be dropped
    # onto FSLeyes to open them
    dt = fsleyesframe.OverlayDropTarget(overlayList, displayCtx)
    frame.SetDropTarget(dt)

    # Make sure the new frame is shown
    # before destroying the splash screen
    frame.Show(True)
    frame.Refresh()
    frame.Update()

    # In certain instances under Linux/GTK,
    # closing the splash screen will crash
    # the application. No idea why. So we
    # leave the splash screen hidden, but
    # not closed, and close it when the main
    # frame is closed. This also works under
    # OSX.
    splash.Hide()
    splash.Refresh()
    splash.Update()

    def onFrameDestroy(ev):
        ev.Skip()

        # splash screen may already
        # have been destroyed
        try:              splash.Close()
        except Exception: pass

    frame.Bind(wx.EVT_WINDOW_DESTROY, onFrameDestroy)

    status.update('Setting up scene...')

    # Set the default SceneOpts.performance
    # level so that all created SceneOpts
    # instances will default to it
    if namespace.performance is not None:
        fsldisplay.SceneOpts.performance.setAttribute(
            None, 'default', namespace.performance)

    # If a layout has been specified,
    # we load the layout
    if namespace.scene is not None:
        layouts.loadLayout(frame, namespace.scene)

    # Apply any view-panel specific arguments
    viewPanels = frame.viewPanels
    for viewPanel in viewPanels:

        if not isinstance(viewPanel, canvaspanel.CanvasPanel):
            continue

        displayCtx = viewPanel.displayCtx
        viewOpts   = viewPanel.sceneOpts

        parseargs.applySceneArgs(
            namespace, overlayList, displayCtx, viewOpts)

    # If a script has been specified, we run
    # the script. This has to be done on the
    # idle loop, because overlays specified
    # on the command line are loaded on the
    # idle loop. Therefore, if we schedule the
    # script on idle (which is a queue), the
    # script can assume that all overlays have
    # already been loaded.
    from fsleyes.actions.runscript import RunScriptAction
    if script is not None:
        idle.idle(frame.menuActions[RunScriptAction], script)

    return frame


def fslDirWarning(parent):
    """Checks to see if the ``$FSLDIR`` environment variable is set, or
    if a FSL installation directory has been saved previously. If not,
    displays a warning via a :class:`.FSLDirDialog`.

    :arg parent: A ``wx`` parent object.
    """

    if fslplatform.fsldir is not None:
        return

    import fsl.utils.settings as fslsettings

    # Check settings before
    # prompting the user
    fsldir = fslsettings.read('fsldir')

    if fsldir is not None:
        fslplatform.fsldir = fsldir
        return

    from fsleyes_widgets.dialog import FSLDirDialog

    dlg = FSLDirDialog(parent, 'FSLeyes', fslplatform.os == 'Darwin')

    if dlg.ShowModal() == wx.ID_OK:

        fsldir = dlg.GetFSLDir()

        log.debug('Setting $FSLDIR to {} (specified '
                  'by user)'.format(fsldir))

        fslplatform.fsldir        = fsldir
        fslsettings.write('fsldir', fsldir)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# overlaydisplaytoolbar.py - The OverlayDisplayToolBar.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>

"""This module provides the :class:`OverlyDisplyaToolBar`, a
:class:`.ControlToolBar` containing controls for changing the display settings
of the currently selected overlay.
"""


import            logging
import os.path as op

import wx

import fsleyes_props                  as props
import fsleyes_widgets.utils.typedict as td
import fsleyes.controls.controlpanel  as ctrlpanel
import fsleyes.views.canvaspanel      as canvaspanel
import fsleyes.icons                  as icons
import fsleyes.tooltips               as fsltooltips
import fsleyes.actions                as actions
import fsleyes.colourmaps             as fslcm
import fsleyes.strings                as strings


log = logging.getLogger(__name__)


class OverlayDisplayToolBar(ctrlpanel.ControlToolBar):
    """The ``OverlyDisplyaToolBar`` is a :class:`.ControlToolBar` containing
    controls which allow the user to change the display settings of the
    currently selected overlay (as defined by the
    :attr:`.DisplayContext.selectedOverlay` property). The display settings
    for an overlay are contained in the :class:`.Display` and
    :class:`.DisplayOpts` instances that are associated with that overlay.


    An ``OverlyDisplyaToolBar`` looks something like the following:

    .. image:: images/overlaydisplaytoolbar.png
       :scale: 50%
       :align: center


    The specific controls which are displayed are defined in the
    :attr:`_TOOLBAR_PROPS` dictionary, and are created by the following
    methods:

    .. autosummary::
       :nosignatures:

       _OverlayDisplayToolBar__makeDisplayTools
       _OverlayDisplayToolBar__makeVolumeOptsTools
       _OverlayDisplayToolBar__makeMaskOptsTools
       _OverlayDisplayToolBar__makeLabelOptsTools
       _OverlayDisplayToolBar__makeVectorOptsTools
       _OverlayDisplayToolBar__makeRGBVectorOptsTools
       _OverlayDisplayToolBar__makeLineVectorOptsTools
       _OverlayDisplayToolBar__makeMeshOptsTools
       _OverlayDisplayToolBar__makeGiftiOptsTools
       _OverlayDisplayToolBar__makeFreesurferOptsTools
       _OverlayDisplayToolBar__makeTensorOptsTools
       _OverlayDisplayToolBar__makeSHOptsTools
       _OverlayDisplayToolBar__makeMIPOptsTools
    """


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``OverlayDisplayToolBar`` is only intended to be added to
        :class:`.OrthoPanel`, :class:`.LightBoxPanel`, or
        :class:`.Scene3DPanel` views.
        """
        return [canvaspanel.CanvasPanel]


    def __init__(self, parent, overlayList, displayCtx, viewPanel):
        """Create an ``OverlyDisplyaToolBar``.

        :arg parent:      The :mod:`wx` parent object.

        :arg overlayList: The :class:`.OverlayList` instance.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg viewPanel:   The :class:`.ViewPanel` which this
                          ``OverlayDisplayToolBar`` is owned by.
        """

        ctrlpanel.ControlToolBar.__init__(self,
                                          parent,
                                          overlayList,
                                          displayCtx,
                                          viewPanel,
                                          height=24,
                                          kbFocus=True)

        self.__viewPanel      = viewPanel
        self.__currentOverlay = None

        self.displayCtx.addListener(
            'selectedOverlay',
            self.name,
            self.__selectedOverlayChanged)
        self.overlayList.addListener(
            'overlays',
            self.name,
            self.__selectedOverlayChanged)

        self.__selectedOverlayChanged()


    def destroy(self):
        """Must be called when this ``OverlyDisplyaToolBar`` is no longer
        needed. Removes some property listeners, and calls the
        :meth:`.ControlToolBar.destroy` method.
        """

        self.overlayList.removeListener('overlays',        self.name)
        self.displayCtx .removeListener('selectedOverlay', self.name)

        if self.__currentOverlay is not None and \
           self.__currentOverlay in self.overlayList:

            display = self.displayCtx.getDisplay(self.__currentOverlay)
            display.removeListener('overlayType', self.name)
            display.removeListener('enabled',     self.name)

        self.__currentOverlay = None
        self.__viewPanel      = None

        ctrlpanel.ControlToolBar.destroy(self)


    def __showTools(self, overlay):
        """Creates and shows a set of controls allowing the user to change
        the display settings of the specified ``overlay``.
        """

        oldTools = self.GetTools()

        # See long comment at bottom
        def destroyOldTools():
            for t in oldTools:
                t.Destroy()

        for t in oldTools:
            t.Show(False)

        self.ClearTools(destroy=False, postevent=False)

        log.debug('Showing tools for {}'.format(overlay))

        display   = self.displayCtx.getDisplay(overlay)
        opts      = display.opts
        tools     = []
        nav       = []

        # Display tools
        dispTools, dispNav = self.__makeDisplayTools(display)

        tools.extend(dispTools)
        nav  .extend(dispNav)

        # DisplayOpts tools
        makeFunc = getattr(self, '_{}__make{}Tools'.format(
            type(self).__name__, type(opts).__name__), None)

        if makeFunc is not None:
            optsTools, optsNav = makeFunc(opts)

            tools.extend(optsTools)
            nav  .extend(optsNav)

        self.SetTools(   tools)
        self.setNavOrder(nav)

        # This method may have been called via an
        # event handler an existing tool in the
        # toolbar - in this situation, destroying
        # that tool will result in nasty crashes,
        # as the wx widget that generated the event
        # will be destroyed while said event is
        # being processed. So we destroy the old
        # tools asynchronously, well after the event
        # which triggered this method call will have
        # returned.
        wx.CallLater(1000, destroyOldTools)


    def __overlayEnableChanged(self, *a):
        """Called when the :attr:`.Display.enabled` property for the currently
        selected overlay changes. Enables/disables this
        ``OverlayDisplayToolBar`` accordingly.
        """
        display = self.displayCtx.getDisplay(self.__currentOverlay)
        self.Enable(display.enabled)


    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay` or
        :class:`.OverlayList` changes. Ensures that controls for the currently
        selected overlay are being shown.
        """

        if self.__currentOverlay is not None and \
           self.__currentOverlay in self.overlayList:
            display = self.displayCtx.getDisplay(self.__currentOverlay)
            display.removeListener('overlayType', self.name)
            display.removeListener('enabled',     self.name)

        overlay = self.displayCtx.getSelectedOverlay()

        self.__currentOverlay = overlay

        if overlay is None:
            self.ClearTools(destroy=True)
            return

        display = self.displayCtx.getDisplay(overlay)

        display.addListener('enabled',
                            self.name,
                            self.__overlayEnableChanged)
        display.addListener('overlayType',
                            self.name,
                            self.__selectedOverlayChanged)

        self.__showTools(overlay)
        self.Enable(display.enabled)


    def __makeDisplayTools(self, display):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.Display` instance.
        """

        viewPanel = self.__viewPanel

        # Display settings
        nameSpec  = _TOOLBAR_PROPS[display, 'name']
        typeSpec  = _TOOLBAR_PROPS[display, 'overlayType']
        alphaSpec = _TOOLBAR_PROPS[display, 'alpha']
        briSpec   = _TOOLBAR_PROPS[display, 'brightness']
        conSpec   = _TOOLBAR_PROPS[display, 'contrast']


        # Buttons which toggle overlay
        # info and display panel
        # This is really hacky. The FSLeyesFrame
        # setattrs a ToggleControlPanelAction for every
        # built-in control to the view panel object,
        # so here we access attributes called
        # "OverlayDisplayPanel" and "OverlayInfoPanel".
        # This will hopefully change in the future.
        panelSpec = actions.ToggleActionButton(
            'OverlayDisplayPanel',
            actionKwargs={'floatPane' : True},
            icon=[icons.findImageFile('gearHighlight24'),
                  icons.findImageFile('gear24')],
            tooltip=fsltooltips.actions[viewPanel, 'OverlayDisplayPanel'])
        infoSpec = actions.ToggleActionButton(
            'OverlayInfoPanel',
            actionKwargs={'floatPane' : True},
            icon=[icons.findImageFile('informationHighlight24'),
                  icons.findImageFile('information24')],
            tooltip=fsltooltips.actions[viewPanel, 'OverlayInfoPanel'])

        # Name/overlay type and brightness/contrast
        # are respectively placed together
        nameTypePanel = wx.Panel(self)
        briconPanel   = wx.Panel(self)
        nameTypeSizer = wx.BoxSizer(wx.VERTICAL)
        briconSizer   = wx.FlexGridSizer(2, 2, 0, 0)

        briconSizer.AddGrowableCol(1)

        nameTypePanel.SetSizer(nameTypeSizer)
        briconPanel  .SetSizer(briconSizer)

        panelWidget = props.buildGUI(self,          viewPanel, panelSpec)
        infoWidget  = props.buildGUI(self,          viewPanel, infoSpec)
        nameWidget  = props.buildGUI(nameTypePanel, display,   nameSpec)
        typeWidget  = props.buildGUI(nameTypePanel, display,   typeSpec)
        briWidget   = props.buildGUI(briconPanel,   display,   briSpec)
        conWidget   = props.buildGUI(briconPanel,   display,   conSpec)
        alphaWidget = props.buildGUI(self,          display,   alphaSpec)

        briLabel    = wx.StaticText(briconPanel)
        conLabel    = wx.StaticText(briconPanel)

        briLabel.SetLabel(strings.properties[display, 'brightness'])
        conLabel.SetLabel(strings.properties[display, 'contrast'])

        # name/type panel
        nameTypeSizer.Add(nameWidget, flag=wx.EXPAND)
        nameTypeSizer.Add(typeWidget, flag=wx.EXPAND)

        # opacity is given a label
        alphaPanel = self.MakeLabelledTool(
            alphaWidget, strings.properties[display, 'alpha'])

        # bricon panel
        briconSizer.Add(briLabel)
        briconSizer.Add(briWidget)
        briconSizer.Add(conLabel)
        briconSizer.Add(conWidget)

        tools = [panelWidget,
                 infoWidget,
                 nameTypePanel,
                 alphaPanel,
                 briconPanel]
        nav   = [panelWidget,
                 infoWidget,
                 nameWidget,
                 typeWidget,
                 alphaWidget,
                 briWidget,
                 conWidget]

        return tools, nav


    def __makeVolumeOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.VolumeOpts` instance.
        """
        rangeSpec      = _TOOLBAR_PROPS[opts, 'displayRange']
        resetSpec      = _TOOLBAR_PROPS[opts, 'resetDisplayRange']
        cmapSpec       = _TOOLBAR_PROPS[opts, 'cmap']
        negCmapSpec    = _TOOLBAR_PROPS[opts, 'negativeCmap']
        useNegCmapSpec = _TOOLBAR_PROPS[opts, 'useNegativeCmap']

        cmapPanel = wx.Panel(self)

        rangeWidget      = props.buildGUI(self,      opts, rangeSpec)
        resetWidget      = props.buildGUI(self,      opts, resetSpec)
        useNegCmapWidget = props.buildGUI(self,      opts, useNegCmapSpec)
        cmapWidget       = props.buildGUI(cmapPanel, opts, cmapSpec)
        negCmapWidget    = props.buildGUI(cmapPanel, opts, negCmapSpec)

        cmapSizer = wx.BoxSizer(wx.VERTICAL)
        cmapPanel.SetSizer(cmapSizer)
        cmapSizer.Add(cmapWidget)
        cmapSizer.Add(negCmapWidget)

        tools = [resetWidget, rangeWidget, useNegCmapWidget, cmapPanel]
        nav   = [resetWidget, rangeWidget, useNegCmapWidget, cmapWidget,
                 negCmapWidget]

        return tools, nav


    def __makeComplexOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.ComplexOpts` instance.
        """
        rangeSpec = _TOOLBAR_PROPS[opts, 'displayRange']
        resetSpec = _TOOLBAR_PROPS[opts, 'resetDisplayRange']
        cmapSpec  = _TOOLBAR_PROPS[opts, 'cmap']
        compSpec  = _TOOLBAR_PROPS[opts, 'component']
        ccpanel   = wx.Panel(self)

        rangeWidget = props.buildGUI(self,    opts, rangeSpec)
        resetWidget = props.buildGUI(self,    opts, resetSpec)
        cmapWidget  = props.buildGUI(ccpanel, opts, cmapSpec)
        compWidget  = props.buildGUI(ccpanel, opts, compSpec)

        ccsizer = wx.BoxSizer(wx.VERTICAL)
        ccpanel.SetSizer(ccsizer)
        ccsizer.Add(cmapWidget, flag=wx.EXPAND)
        ccsizer.Add(compWidget, flag=wx.EXPAND)

        tools = [resetWidget, rangeWidget, ccpanel]
        nav   = [resetWidget, rangeWidget, cmapWidget, compWidget]

        return tools, nav


    def __makeMaskOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.MaskOpts` instance.
        """
        thresSpec  = _TOOLBAR_PROPS[opts, 'threshold']
        colourSpec = _TOOLBAR_PROPS[opts, 'colour']

        thresWidget  = props.buildGUI(self, opts, thresSpec)
        colourWidget = props.buildGUI(self, opts, colourSpec)

        tools = [thresWidget, colourWidget]

        return tools, tools


    def __makeLabelOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.LabelOpts` instance.
        """

        lutSpec     = _TOOLBAR_PROPS[opts, 'lut']
        outlineSpec = _TOOLBAR_PROPS[opts, 'outline']
        widthSpec   = _TOOLBAR_PROPS[opts, 'outlineWidth']

        # lut/outline width widgets
        # are on a single panel
        lutWidthPanel = wx.Panel(self)
        lutWidthSizer = wx.FlexGridSizer(2, 2, 0, 0)
        lutWidthPanel.SetSizer(lutWidthSizer)

        lutWidget     = props.buildGUI(lutWidthPanel, opts, lutSpec)
        widthWidget   = props.buildGUI(lutWidthPanel, opts, widthSpec)
        outlineWidget = props.buildGUI(self,          opts, outlineSpec)

        lutLabel   = wx.StaticText(lutWidthPanel)
        widthLabel = wx.StaticText(lutWidthPanel)

        lutLabel  .SetLabel(strings.properties[opts, 'lut'])
        widthLabel.SetLabel(strings.properties[opts, 'outlineWidth'])

        lutWidthSizer.Add(lutLabel)
        lutWidthSizer.Add(lutWidget,   flag=wx.EXPAND)
        lutWidthSizer.Add(widthLabel)
        lutWidthSizer.Add(widthWidget, flag=wx.EXPAND)

        tools = [lutWidthPanel, outlineWidget]
        nav   = [lutWidget, widthWidget, outlineWidget]

        return tools, nav


    def __makeVectorOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.VectorOpts` instance.
        """

        modSpec   = _TOOLBAR_PROPS[opts, 'modulateImage']
        clipSpec  = _TOOLBAR_PROPS[opts, 'clipImage']
        rangeSpec = _TOOLBAR_PROPS[opts, 'clippingRange']

        panel = wx.Panel(self)
        sizer = wx.GridBagSizer()
        panel.SetSizer(sizer)

        modWidget   = props.buildGUI(panel, opts, modSpec)
        clipWidget  = props.buildGUI(panel, opts, clipSpec)
        rangeWidget = props.buildGUI(panel, opts, rangeSpec)
        modLabel    = wx.StaticText(panel)
        clipLabel   = wx.StaticText(panel)

        modLabel .SetLabel(strings.properties[opts, 'modulateImage'])
        clipLabel.SetLabel(strings.properties[opts, 'clipImage'])

        sizer.Add(modLabel,    pos=(0, 0), span=(1, 1), flag=wx.EXPAND)
        sizer.Add(modWidget,   pos=(0, 1), span=(1, 1), flag=wx.EXPAND)
        sizer.Add(clipLabel,   pos=(1, 0), span=(1, 1), flag=wx.EXPAND)
        sizer.Add(clipWidget,  pos=(1, 1), span=(1, 1), flag=wx.EXPAND)
        sizer.Add(rangeWidget, pos=(0, 2), span=(2, 1), flag=wx.EXPAND)

        tools = [panel]
        nav   = [modWidget, clipWidget, rangeWidget]

        return tools, nav


    def __makeRGBVectorOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.RGBVectorOpts` instance.
        """
        return self.__makeVectorOptsTools(opts)


    def __makeLineVectorOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.LineVectorOpts` instance.
        """
        widthSpec = _TOOLBAR_PROPS[opts, 'lineWidth']

        widget    = props.buildGUI(self, opts, widthSpec)
        lblWidget = self.MakeLabelledTool(
            widget, strings.properties[opts, 'lineWidth'])

        tools, nav = self.__makeVectorOptsTools(opts)

        return tools + [lblWidget], nav + [widget]


    def __makeMeshOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.MeshOpts` instance.
        """
        colourSpec  = _TOOLBAR_PROPS[opts, 'colour']
        outlineSpec = _TOOLBAR_PROPS[opts, 'outline']
        widthSpec   = _TOOLBAR_PROPS[opts, 'outlineWidth']

        colourWidget  = props.buildGUI(self, opts, colourSpec)
        outlineWidget = props.buildGUI(self, opts, outlineSpec)
        widthWidget   = props.buildGUI(self, opts, widthSpec)

        lblWidthWidget  = self.MakeLabelledTool(
            widthWidget, strings.properties[opts, 'outlineWidth'])

        tools = [colourWidget, outlineWidget, lblWidthWidget]
        nav   = [colourWidget, outlineWidget, widthWidget]

        return tools, nav


    def __makeGiftiOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.GiftiOpts` instance.
        """
        tools, nav = self.__makeMeshOptsTools(opts)

        vertWidget  = _TOOLBAR_PROPS[opts, 'vertexSet']
        vdataWidget = _TOOLBAR_PROPS[opts, 'vertexData']

        panel       = wx.Panel(self)
        sizer       = wx.BoxSizer(wx.VERTICAL)
        vertWidget  = props.buildGUI(panel, opts, vertWidget)
        vdataWidget = props.buildGUI(panel, opts, vdataWidget)

        sizer.Add(vertWidget,  flag=wx.EXPAND)
        sizer.Add(vdataWidget, flag=wx.EXPAND)
        panel.SetSizer(sizer)

        tools += [panel]
        nav   += [vertWidget, vdataWidget]

        return tools, nav


    def __makeFreesurferOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.FreesurferOpts` instance.
        """
        return self.__makeGiftiOptsTools(opts)


    def __makeTensorOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.TensorOpts` instance.
        """
        lightingSpec   = _TOOLBAR_PROPS[opts, 'lighting']
        lightingWidget = props.buildGUI(self, opts, lightingSpec)

        tools, nav = self.__makeVectorOptsTools(opts)

        return tools + [lightingWidget], nav + [lightingWidget]


    def __makeSHOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.SHOpts` instance.
        """
        sizeSpec = _TOOLBAR_PROPS[opts, 'size']
        radSpec  = _TOOLBAR_PROPS[opts, 'radiusThreshold']

        panel = wx.Panel(self)
        sizer = wx.FlexGridSizer(2, 2, 0, 0)

        sizeWidget  = props.buildGUI(panel, opts, sizeSpec)
        radWidget   = props.buildGUI(panel, opts, radSpec)

        sizeLabel = wx.StaticText(panel)
        radLabel  = wx.StaticText(panel)

        sizeLabel.SetLabel(strings.properties[opts, 'size'])
        radLabel .SetLabel(strings.properties[opts, 'radiusThreshold'])

        sizer.Add(sizeLabel)
        sizer.Add(sizeWidget)
        sizer.Add(radLabel)
        sizer.Add(radWidget)
        panel.SetSizer(sizer)

        tools, nav = self.__makeVectorOptsTools(opts)

        tools.extend([panel])
        nav.extend(  [sizeWidget, radWidget])

        return tools, nav


    def __makeMIPOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.MIPOpts` instance.
        """
        rangeSpec = _TOOLBAR_PROPS[opts, 'displayRange']
        resetSpec = _TOOLBAR_PROPS[opts, 'resetDisplayRange']
        cmapSpec  = _TOOLBAR_PROPS[opts, 'cmap']

        cmapPanel = wx.Panel(self)

        rangeWidget = props.buildGUI(self,      opts, rangeSpec)
        resetWidget = props.buildGUI(self,      opts, resetSpec)
        cmapWidget  = props.buildGUI(cmapPanel, opts, cmapSpec)

        cmapSizer = wx.BoxSizer(wx.VERTICAL)
        cmapPanel.SetSizer(cmapSizer)
        cmapSizer.Add(cmapWidget)

        tools = [resetWidget, rangeWidget, cmapPanel]
        nav   = [resetWidget, rangeWidget, cmapWidget]

        return tools, nav


def _imageLabel(img):
    """Used to generate labels for the :attr:`.VectorOpts.modulateImage`,
    :attr:`.VectorOpts.clipImage`, and other :class:`.Image`-based
    choice properties.
    """
    if img is None: return 'None'
    else:           return img.name



def _pathLabel(p):
    if p is None: return 'None'
    else:         return op.basename(p)


_TOOLTIPS = td.TypeDict({

    'Display.name'        : fsltooltips.properties['Display.name'],
    'Display.overlayType' : fsltooltips.properties['Display.overlayType'],
    'Display.alpha'       : fsltooltips.properties['Display.alpha'],
    'Display.brightness'  : fsltooltips.properties['Display.brightness'],
    'Display.contrast'    : fsltooltips.properties['Display.contrast'],

    'ColourMapOpts.displayRange' :
    fsltooltips.properties['ColourMapOpts.displayRange'],
    'VolumeOpts.cmap'              :
    fsltooltips.properties['ColourMapOpts.cmap'],
    'VolumeOpts.negativeCmap'      :
    fsltooltips.properties['ColourMapOpts.negativeCmap'],
    'VolumeOpts.useNegativeCmap'   :
    fsltooltips.properties['ColourMapOpts.useNegativeCmap'],

    'ComplexOpts.component' : fsltooltips.properties['ComplexOpts.component'],

    'MaskOpts.threshold' : fsltooltips.properties['MaskOpts.threshold'],
    'MaskOpts.colour'    : fsltooltips.properties['MaskOpts.colour'],

    'LabelOpts.lut'          : fsltooltips.properties['LabelOpts.lut'],

    'LabelOpts.outline'      : fsltooltips.properties['LabelOpts.outline'],
    'LabelOpts.outlineWidth' : fsltooltips.properties['LabelOpts.'
                                                      'outlineWidth'],

    'VectorOpts.modulateImage' : fsltooltips.properties['VectorOpts.'
                                                        'modulateImage'],
    'VectorOpts.clipImage'     : fsltooltips.properties['VectorOpts.'
                                                        'clipImage'],
    'VectorOpts.clippingRange' : fsltooltips.properties['VectorOpts.'
                                                        'clippingRange'],

    'LineVectorOpts.lineWidth' : fsltooltips.properties['LineVectorOpts.'
                                                        'lineWidth'],

    'MeshOpts.colour'       : fsltooltips.properties['MeshOpts.colour'],
    'MeshOpts.outline'      : fsltooltips.properties['MeshOpts.outline'],
    'MeshOpts.outlineWidth' : fsltooltips.properties['MeshOpts.'
                                                     'outlineWidth'],
    'MeshOpts.vertexSet'    : fsltooltips.properties['MeshOpts.vertexSet'],
    'MeshOpts.vertexData'   : fsltooltips.properties['MeshOpts.vertexData'],

    'TensorOpts.lighting' : fsltooltips.properties['TensorOpts.'
                                                   'lighting'],

    'SHOpts.lighting'        : fsltooltips.properties['SHOpts.lighting'],
    'SHOpts.size'            : fsltooltips.properties['SHOpts.size'],
    'SHOpts.radiusThreshold' : fsltooltips.properties['SHOpts.'
                                                      'radiusThreshold'],
})
"""This dictionary contains tooltips for :class:`.Display` and
:class:`.DisplayOpts` properties. It is referenced in the
:attr:`_TOOLBAR_PROPS` dictionary definition.
"""


_TOOLBAR_PROPS = td.TypeDict({

    'Display.name'         : props.Widget(
        'name',
        tooltip=_TOOLTIPS['Display.name']),
    'Display.overlayType'  : props.Widget(
        'overlayType',
        tooltip=_TOOLTIPS['Display.overlayType'],
        labels=strings.choices['Display.overlayType']),
    'Display.alpha'        : props.Widget(
        'alpha',
        spin=False,
        showLimits=False,
        tooltip=_TOOLTIPS['Display.alpha']),
    'Display.brightness'   : props.Widget(
        'brightness',
        spin=False,
        showLimits=False,
        tooltip=_TOOLTIPS['Display.brightness']),
    'Display.contrast'     : props.Widget(
        'contrast',
        spin=False,
        showLimits=False,
        tooltip=_TOOLTIPS['Display.contrast']),

    'VolumeOpts.resetDisplayRange' : actions.ActionButton(
        'resetDisplayRange',
        icon=icons.findImageFile('verticalReset32')
    ),

    'VolumeOpts.displayRange' : props.Widget(
        'displayRange',
        slider=False,
        showLimits=False,
        spinWidth=10,
        tooltip=_TOOLTIPS['ColourMapOpts.displayRange'],
        labels=[strings.choices['ColourMapOpts.displayRange.min'],
                strings.choices['ColourMapOpts.displayRange.max']]),
    'VolumeOpts.cmap' : props.Widget(
        'cmap',
        labels=fslcm.getColourMapLabel,
        tooltip=_TOOLTIPS['VolumeOpts.cmap']),
    'VolumeOpts.useNegativeCmap' : props.Widget(
        'useNegativeCmap',
        icon=[icons.findImageFile('twocmaps24'),
              icons.findImageFile('onecmap24')],
        toggle=True,
        tooltip=_TOOLTIPS['VolumeOpts.useNegativeCmap']),

    'VolumeOpts.negativeCmap' : props.Widget(
        'negativeCmap',
        labels=fslcm.getColourMapLabel,
        tooltip=_TOOLTIPS['VolumeOpts.negativeCmap'],
        dependencies=['useNegativeCmap'],
        enabledWhen=lambda i, unc : unc),

    'ComplexOpts.component' : props.Widget(
        'component',
        labels=strings.choices['ComplexOpts.component'],
        tooltip=_TOOLTIPS['ComplexOpts.component']),

    'MaskOpts.threshold' : props.Widget(
        'threshold',
        showLimits=False,
        spin=True,
        tooltip=_TOOLTIPS['MaskOpts.threshold'],
        labels=[strings.choices['MaskOpts.threshold.min'],
                strings.choices['MaskOpts.threshold.max']]),
    'MaskOpts.colour'    : props.Widget(
        'colour',
        size=(24, 24),
        tooltip=_TOOLTIPS['MaskOpts.colour']),

    'LabelOpts.lut'     : props.Widget(
        'lut',
        tooltip=_TOOLTIPS['LabelOpts.lut'],
        labels=lambda l: l.name),
    'LabelOpts.outline' : props.Widget(
        'outline',
        tooltip=_TOOLTIPS['LabelOpts.outline'],
        icon=[icons.findImageFile('outline24'),
              icons.findImageFile('filled24')],
        toggle=True),
    'LabelOpts.outlineWidth' : props.Widget(
        'outlineWidth',
        tooltip=_TOOLTIPS['LabelOpts.outlineWidth'],
        showLimits=False,
        spin=False),

    'MeshOpts.colour'       : props.Widget(
        'colour',
        size=(24, 24),
        tooltip=_TOOLTIPS['MeshOpts.colour']),
    'MeshOpts.outline'      : props.Widget(
        'outline',
        tooltip=_TOOLTIPS['MeshOpts.outline'],
        icon=[icons.findImageFile('outline24'),
              icons.findImageFile('filled24')],
        toggle=True),
    'MeshOpts.outlineWidth' : props.Widget(
        'outlineWidth',
        showLimits=False,
        spin=False,
        tooltip=_TOOLTIPS['MeshOpts.outlineWidth'],
        enabledWhen=lambda i: i.outline),
    'MeshOpts.vertexSet' : props.Widget(
        'vertexSet',
        labels=_pathLabel,
        tooltip=_TOOLTIPS['MeshOpts.vertexSet']
    ),
    'MeshOpts.vertexData' : props.Widget(
        'vertexData',
        labels=_pathLabel,
        tooltip=_TOOLTIPS['MeshOpts.vertexData']
    ),

    'VectorOpts.modulateImage' : props.Widget(
        'modulateImage',
        labels=_imageLabel,
        tooltip=_TOOLTIPS['VectorOpts.modulateImage']),
    'VectorOpts.clipImage' : props.Widget(
        'clipImage',
        labels=_imageLabel,
        tooltip=_TOOLTIPS['VectorOpts.clipImage']),
    'VectorOpts.clippingRange' : props.Widget(
        'clippingRange',
        showLimits=False,
        slider=True,
        spin=False,
        tooltip=_TOOLTIPS['VectorOpts.clippingRange'],
        labels=[strings.choices['VectorOpts.clippingRange.min'],
                strings.choices['VectorOpts.clippingRange.max']],
        dependencies=['clipImage'],
        enabledWhen=lambda o, ci: ci is not None),

    'LineVectorOpts.lineWidth' : props.Widget(
        'lineWidth',
        showLimits=False,
        spin=False,
        tooltip=_TOOLTIPS['LineVectorOpts.lineWidth']),

    'TensorOpts.lighting'      : props.Widget(
        'lighting',
        icon=[icons.findImageFile('lightbulbHighlight24'),
              icons.findImageFile('lightbulb24')],
        tooltip=_TOOLTIPS['TensorOpts.lighting']),

    'SHOpts.lighting'        : props.Widget(
        'lighting',
        icon=[icons.findImageFile('lightbulbHighlight24'),
              icons.findImageFile('lightbulb24')],
        tooltip=_TOOLTIPS['SHOpts.lighting']),
    'SHOpts.size'            : props.Widget(
        'size',
        showLimits=False,
        slider=True,
        spin=False,
        tooltip=_TOOLTIPS['SHOpts.size']),
    'SHOpts.radiusThreshold' : props.Widget(
        'radiusThreshold',
        slider=True,
        spin=False,
        showLimits=False,
        tooltip=_TOOLTIPS['SHOpts.radiusThreshold']),


    'MIPOpts.resetDisplayRange' : actions.ActionButton(
        'resetDisplayRange',
        icon=icons.findImageFile('verticalReset32')
    ),

    'MIPOpts.displayRange' : props.Widget(
        'displayRange',
        slider=False,
        showLimits=False,
        spinWidth=10,
        tooltip=_TOOLTIPS['ColourMapOpts.displayRange'],
        labels=[strings.choices['ColourMapOpts.displayRange.min'],
                strings.choices['ColourMapOpts.displayRange.max']]),
    'MIPOpts.cmap' : props.Widget(
        'cmap',
        labels=fslcm.getColourMapLabel,
        tooltip=_TOOLTIPS['VolumeOpts.cmap']),
})
"""This dictionary defines specifications for all controls shown on an
:class:`OverlayDisplayToolBar`.
"""



#!/usr/bin/env python
#
# overlaydisplaypanel.py - The OverlayDisplayPanel.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>

"""This module provides the :class:`OverlayDisplayPanel` class, a *FSLeyes
control* panel which allows the user to change overlay display settings.
"""


import logging
import functools
import collections
import collections.abc as abc

import wx

import fsleyes_props                  as props
import fsleyes.views.canvaspanel      as canvaspanel
import fsleyes.controls.controlpanel  as ctrlpanel
import fsleyes.strings                as strings
import fsleyes.tooltips               as fsltooltips

from . import overlaydisplaywidgets   as odwidgets


log = logging.getLogger(__name__)


class OverlayDisplayPanel(ctrlpanel.SettingsPanel):
    """The ``OverlayDisplayPanel`` is a :class:`.SettingsPanel` which allows
    the user to change the display settings of the currently selected
    overlay (which is defined by the :attr:`.DisplayContext.selectedOverlay`
    property). The display settings for an overlay are contained in the
    :class:`.Display` and :class:`.DisplayOpts` instances associated with
    that overlay. An ``OverlayDisplayPanel`` looks something like the
    following:

    .. image:: images/overlaydisplaypanel.png
       :scale: 50%
       :align: center

    An ``OverlayDisplayPanel`` uses a :class:`.WidgetList` to organise the
    settings into two main sections:

      - Settings which are common across all overlays - these are defined
        in the :class:`.Display` class.

      - Settings which are specific to the current
        :attr:`.Display.overlayType` - these are defined in the
        :class:`.DisplayOpts` sub-classes.


    The settings that are displayed on an ``OverlayDisplayPanel`` are
    defined in the :attr:`_DISPLAY_PROPS` and :attr:`_DISPLAY_WIDGETS`
    dictionaries.
    """


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``OverlayDisplayPanel`` is only intended to be added to
        :class:`.OrthoPanel`, :class:`.LightBoxPanel`, or
        :class:`.Scene3DPanel` views.
        """
        return [canvaspanel.CanvasPanel]


    @staticmethod
    def defaultLayout():
        """Returns a dictionary containing layout settings to be passed to
        :class:`.ViewPanel.togglePanel`.
        """
        return {'location' : wx.LEFT}


    def __init__(self, parent, overlayList, displayCtx, canvasPanel):
        """Create an ``OverlayDisplayPanel``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg canvasPanel: The :class:`.CanvasPanel` instance.
        """

        from fsleyes.views.scene3dpanel import Scene3DPanel

        ctrlpanel.SettingsPanel.__init__(self,
                                         parent,
                                         overlayList,
                                         displayCtx,
                                         canvasPanel,
                                         kbFocus=True)

        displayCtx .addListener('selectedOverlay',
                                 self.name,
                                 self.__selectedOverlayChanged)
        overlayList.addListener('overlays',
                                 self.name,
                                 self.__selectedOverlayChanged)

        self.__threedee       = isinstance(parent, Scene3DPanel)
        self.__viewPanel      = canvasPanel
        self.__widgets        = None
        self.__currentOverlay = None

        self.__selectedOverlayChanged()


    def destroy(self):
        """Must be called when this ``OverlayDisplayPanel`` is no longer
        needed. Removes property listeners, and calls the
        :meth:`.SettingsPanel.destroy` method.
        """

        self.displayCtx .removeListener('selectedOverlay', self.name)
        self.overlayList.removeListener('overlays',        self.name)

        if self.__currentOverlay is not None and \
           self.__currentOverlay in self.overlayList:

            display = self.displayCtx.getDisplay(self.__currentOverlay)

            display.removeListener('overlayType', self.name)

        self.__viewPanel      = None
        self.__widgets        = None
        self.__currentOverlay = None

        ctrlpanel.SettingsPanel.destroy(self)


    def __selectedOverlayChanged(self, *a):
        """Called when the :class:`.OverlayList` or
        :attr:`.DisplayContext.selectedOverlay` changes. Refreshes this
        ``OverlayDisplayPanel`` so that the display settings for the newly
        selected overlay are shown.
        """

        overlay     = self.displayCtx.getSelectedOverlay()
        lastOverlay = self.__currentOverlay
        widgetList  = self.getWidgetList()

        if overlay is None:
            self.__currentOverlay = None
            self.__widgets        = None
            widgetList.Clear()
            self.Layout()
            return

        if overlay is lastOverlay:
            return

        self.__currentOverlay = overlay
        self.__widgets        = collections.OrderedDict()

        display = self.displayCtx.getDisplay(overlay)
        opts    = display.opts

        if self.__threedee:
            groups  = ['display', 'opts', '3d']
            targets = [ display,   opts,   opts]
            labels  = [strings.labels[self, display],
                       strings.labels[self, opts],
                       strings.labels[self, '3d']]

        else:
            groups  = ['display', 'opts']
            targets = [ display,   opts]
            labels  = [strings.labels[self, display],
                       strings.labels[self, opts]]

        keepExpanded = {g : True for g in groups}

        if lastOverlay is not None and lastOverlay in self.overlayList:

            lastDisplay = self.displayCtx.getDisplay(lastOverlay)
            lastDisplay.removeListener('overlayType', self.name)

        if lastOverlay is not None:
            for g in groups:
                keepExpanded[g] = widgetList.IsExpanded(g)

        display.addListener('overlayType', self.name, self.__ovlTypeChanged)

        widgetList.Clear()

        for g, l, t in zip(groups, labels, targets):

            widgetList.AddGroup(g, l)
            self.__widgets[g] = self.__updateWidgets(t, g)
            widgetList.Expand(g, keepExpanded[g])

        self.setNavOrder()
        self.Layout()


    def setNavOrder(self):

        allWidgets = self.__widgets.items()
        allWidgets = functools.reduce(lambda a, b: a + b, allWidgets)

        ctrlpanel.SettingsPanel.setNavOrder(self, allWidgets)


    def __ovlTypeChanged(self, *a):
        """Called when the :attr:`.Display.overlayType` of the current overlay
        changes. Refreshes the :class:`.DisplayOpts` settings which are shown,
        as a new :class:`.DisplayOpts` instance will have been created for the
        overlay.
        """

        opts       = self.displayCtx.getOpts(self.__currentOverlay)
        widgetList = self.getWidgetList()

        self.__widgets[opts] = self.__updateWidgets(opts, 'opts')

        widgetList.RenameGroup('opts', strings.labels[self, opts])

        if '3d' in self.__widgets:
            self.__widgets['3d'] = self.__updateWidgets(opts, '3d')

        self.setNavOrder()
        self.Layout()


    def updateWidgets(self, target, groupName):
        """Re-generates the widgets for the given target/group. """

        self.__widgets[target] = self.__updateWidgets(target, groupName)
        self.setNavOrder()
        self.Layout()


    def __updateWidgets(self, target, groupName):
        """Called by the :meth:`__selectedOverlayChanged` and
        :meth:`__ovlTypeChanged` methods. Re-creates the controls on this
        ``OverlayDisplayPanel`` for the specified group.

        :arg target:    A :class:`.Display` or :class:`.DisplayOpts` instance,
                        which contains the properties that controls are to be
                        created for.

        :arg groupName: Either ``'display'`` or ``'opts'``/``'3d'``,
                        corresponding to :class:`.Display` or
                        :class:`.DisplayOpts` properties.

        :returns:       A list containing all of the new widgets that
                        were created.
        """

        widgetList = self.getWidgetList()

        widgetList.ClearGroup(groupName)

        if groupName == '3d':
            dispProps = odwidgets.get3DPropertyList(target)
            dispSpecs = odwidgets.get3DWidgetSpecs( target)
        else:
            dispProps = odwidgets.getPropertyList(target, self.__threedee)
            dispSpecs = odwidgets.getWidgetSpecs( target, self.__threedee)

        allLabels     = []
        allTooltips   = []
        allWidgets    = []
        allContainers = []

        for p in dispProps:

            spec  = dispSpecs[p]

            specs    = [spec]
            labels   = [strings    .properties.get((target, p), None)]
            tooltips = [fsltooltips.properties.get((target, p), None)]

            if callable(spec):

                # Will either return a contsiner
                # widget/sizer and a list of widgets
                # for setting the navigation order,
                # or will return a list of specs
                # (with an irrelevant second parameter)
                container, widgets = spec(
                    target,
                    widgetList,
                    self,
                    self.overlayList,
                    self.displayCtx,
                    self.__threedee)

                if isinstance(container, abc.Sequence):
                    specs    = container
                    keys     = [s.key for s in specs]
                    labels   = [strings.properties.get((target, k), None)
                                for k in keys]
                    tooltips = [fsltooltips.properties.get((target, k), None)
                                for k in keys]

                else:
                    allContainers.append(container)
                    allWidgets   .extend(widgets)
                    specs = []

            for s in specs:
                widget = props.buildGUI(widgetList, target, s)

                allWidgets   .append(widget)
                allContainers.append(widget)

            allLabels  .extend(labels)
            allTooltips.extend(tooltips)

        for widget, label, tooltip in zip(allContainers,
                                          allLabels,
                                          allTooltips):
            if label is None:
                label = ''
            widgetList.AddWidget(
                widget,
                label,
                tooltip=tooltip,
                groupName=groupName)

        return allWidgets



#!/usr/bin/env python
#
# plotcanvas.py - The PlotCanvas class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`PlotCanvas` class, which plots
:class:`.DataSeries` instances on a ``matplotlib`` canvas. The ``PlotCanvas``
is used by the :class:`.TimeSeriesPanel`, :class:`.HistogramPanel`, and
:class:`.PowerSpectrumPanel` views, and potentially by other FSLeyes control
panels.
"""


import logging
import collections

import numpy                             as np
import scipy.interpolate                 as interp
import matplotlib.pyplot                 as plt
import matplotlib.backends.backend_wxagg as wxagg

import fsl.utils.idle                    as idle
import fsleyes_props                     as props
import fsleyes_widgets                   as fwidgets

import fsleyes.strings                   as strings


log = logging.getLogger(__name__)


class PlotCanvas(props.HasProperties):
    """The ``PlotCanvas`` can be used to plot :class:`.DataSeries` instances
    onto a ``matplotlib`` ``FigureCanvasWxAgg`` canvas.

    The ``PlotCanvas`` is used by all *FSLeyes views* which display some sort
    of 2D data plot, such as the :class:`.TimeSeriesPanel`, and the
    :class:`.HistogramPanel`.

    The ``PlotCanvas`` uses :mod:`matplotlib` for its plotting. The
    ``matplotlib`` ``Figure``, ``Axis``, and ``Canvas`` instances can be
    accessed via the :meth:`figure`, :meth:`axis`, and :meth:`canvas` methods,
    if they are needed. Various display settings can be configured through
    ``PlotCanvas`` properties, including :attr:`legend`, :attr:`smooth`, etc.


    **Basic usage**

    After you have created a ``PlotCanvas``, you can add :class:`.DataSeries`
    instances to the :attr:`dataSeries` property, then call :meth:`draw` or
    :meth:`asyncDraw`.

    The :meth:`draw` method simply calls :meth:`drawDataSeries` and
    :meth:`drawArtists`, so you can alternately call those methods directly,
    or pass your own  ``drawFunc`` when creating a ``PlotCanvas``.

    The ``PlotCanvas`` itself is not a ``wx`` object, so cannot be displayed -
    the ``matplotilb Canvas`` object, accessible through the :meth:`canvas`
    method, is what you should add to a ``wx`` parent object.


    **Data series**

    A ``PlotCanvas`` instance plots data contained in one or more
    :class:`.DataSeries` instances; all ``DataSeries`` classes are defined in
    the :mod:`.plotting` sub-package.

    ``DataSeries`` objects can be plotted by passing them to the
    :meth:`drawDataSeries` method.

    Or, if you want one or more ``DataSeries`` to be *held*, i.e. plotted
    every time, you can add them to the :attr:`dataSeries` list. The
    ``DataSeries`` in the :attr:`dataSeries` list will be plotted on every
    call to :meth:`drawDataSeries` (in addition to any ``DataSeries`` passed
    directly to :meth:`drawDataSeries`) until they are removed from the
    :attr:`dataSeries` list.


    **The draw queue**


    The ``PlotCanvas`` uses a :class:`.async.TaskThread` to asynchronously
    extract and prepare data for plotting, This is because data preparation
    may take a long time for certain types of ``DataSeries``
    (e.g. :class:`TimeSeries` which are retrieving data from large
    :class:`.Image` overlays), and the main application thread should not be
    blocked while this is occurring. The ``TaskThread`` instance is accessible
    through the :meth:`getDrawQueue` method, in case anything needs to be
    scheduled on it.
    """


    dataSeries = props.List()
    """This list contains :class:`.DataSeries` instances which are plotted
    on every call to :meth:`drawDataSeries`. ``DataSeries`` instances can
    be added/removed directly to/from this list.
    """


    artists = props.List()
    """This list contains any ``matplotlib.Artist`` instances which are
    plotted every call to :meth:`drawArtists`.
    """


    legend = props.Boolean(default=True)
    """If ``True``, a legend is added to the plot, with an entry for every
    ``DataSeries`` instance in the :attr:`dataSeries` list.
    """


    xAutoScale = props.Boolean(default=True)
    """If ``True``, the plot :attr:`limits` for the X axis are automatically
    updated to fit all plotted data.
    """


    yAutoScale = props.Boolean(default=True)
    """If ``True``, the plot :attr:`limits` for the Y axis are automatically
    updated to fit all plotted data.
    """


    xLogScale = props.Boolean(default=False)
    """Toggle a :math:`log_{10}` x axis scale. """


    yLogScale = props.Boolean(default=False)
    """Toggle a :math:`log_{10}` y axis scale. """


    invertX = props.Boolean(default=False)
    """Invert the plot along the X axis. """


    invertY = props.Boolean(default=False)
    """Invert the plot along the Y axis. """


    xScale = props.Real(default=1)
    """Scale to apply to the X axis data. """


    yScale = props.Real(default=1)
    """Scale to apply to the Y axis data. """


    xOffset = props.Real(default=0)
    """Offset to apply to the X axis data. """


    yOffset = props.Real(default=0)
    """Offset to apply to the Y axis data. """


    ticks = props.Boolean(default=True)
    """Toggle axis ticks and tick labels on/off."""


    grid = props.Boolean(default=True)
    """Toggle an axis grid on/off."""


    gridColour = props.Colour(default=(1, 1, 1))
    """Grid colour (if :attr:`grid` is ``True``)."""


    bgColour = props.Colour(default=(0.8, 0.8, 0.8))
    """Plot background colour."""


    smooth = props.Boolean(default=False)
    """If ``True`` all plotted data is up-sampled, and smoothed using
    spline interpolation.
    """


    xlabel = props.String()
    """A label to show on the x axis. """


    ylabel = props.String()
    """A label to show on the y axis. """


    limits = props.Bounds(ndims=2)
    """The x/y axis limits. If :attr:`xAutoScale` and :attr:`yAutoScale` are
    ``True``, these limit values are automatically updated on every call to
    :meth:`drawDataSeries`.
    """


    def __init__(self,
                 parent,
                 drawFunc=None,
                 prepareFunc=None):
        """Create a ``PlotCanvas``.

        :arg parent:      The :mod:`wx` parent object.
        :arg drawFunc:    Custon function to call instead of :meth:`draw`.
        :arg prepareFunc: Custom function to call instead of
                          :meth:`prepareDataSeries`.
        """

        figure = plt.Figure()
        axis   = figure.add_subplot(111)
        canvas = wxagg.FigureCanvasWxAgg(parent, -1, figure)

        figure.subplots_adjust(top=1.0, bottom=0.0, left=0.0, right=1.0)
        figure.patch.set_visible(False)

        self.__name        = '{}_{}'.format(type(self).__name__, id(self))
        self.__figure      = figure
        self.__axis        = axis
        self.__canvas      = canvas
        self.__prepareFunc = prepareFunc
        self.__drawFunc    = drawFunc
        self.__destroyed   = False

        # Accessing data from large compressed
        # files may take time, so we maintain
        # a queue of plotting requests. The
        # functions executed on this task
        # thread are used to prepare data for
        # plotting - the plotting occurs on
        # the main WX event loop.
        #
        # The drawDataSeries method sets up the
        # asynchronous data preparation, and the
        # __drawDataSeries method does the actual
        # plotting.
        self.__drawQueue = idle.TaskThread()
        self.__drawQueue.daemon = True
        self.__drawQueue.start()

        # Whenever a new request comes in to
        # draw the plot, we can't cancel any
        # pending requests, as they are running
        # on separate threads and out of our
        # control (and could be blocking on I/O).
        #
        # Instead, we keep track of the total
        # number of pending requests. The
        # __drawDataSeries method (which does the
        # actual plotting) will only draw the
        # plot if there are no pending requests
        # (because otherwise it would be drawing
        # out-of-date data).
        self.__drawRequests = 0

        # The getDrawnDataSeries method returns
        # data as it is shown on the plot - some
        # pre/post-processing may be applied to
        # the data as retrieved by DataSeries
        # instances, so this dictionary is used
        # to keep copies of the mpl Artist object
        # which contains the data with all
        # processing applied, that is currrently
        # on the plot (and accessible via
        # getDrawnDataSeries).
        self.__drawnDataSeries = collections.OrderedDict()

        # Redraw whenever any property changes,
        for propName in ['legend',
                         'xAutoScale',
                         'yAutoScale',
                         'xLogScale',
                         'yLogScale',
                         'invertX',
                         'invertY',
                         'xScale',
                         'yScale',
                         'xOffset',
                         'yOffset',
                         'ticks',
                         'grid',
                         'gridColour',
                         'bgColour',
                         'smooth',
                         'xlabel',
                         'ylabel']:
            self.addListener(propName, self.__name, self.asyncDraw)

        # custom listeners for a couple of properties
        self.addListener('dataSeries',
                         self.__name,
                         self.__dataSeriesChanged)
        self.addListener('artists',
                         self.__name,
                         self.__artistsChanged)
        self.addListener('limits',
                         self.__name,
                         self.__limitsChanged)


    def destroy(self):
        """Removes some property listeners, and clears references to all
        :class:`.DataSeries`, ``matplotlib`` and ``wx`` objects.
        """

        for propName in ['dataSeries',
                         'artists',
                         'limits',
                         'legend',
                         'xAutoScale',
                         'yAutoScale',
                         'xLogScale',
                         'yLogScale',
                         'invertX',
                         'invertY',
                         'xScale',
                         'yScale',
                         'xOffset',
                         'yOffset',
                         'ticks',
                         'grid',
                         'gridColour',
                         'bgColour',
                         'smooth',
                         'xlabel',
                         'ylabel']:
            self.removeListener(propName, self.__name)

        for ds in self.dataSeries:

            for propName in ds.redrawProperties():
                ds.removeListener(propName, self.__name)
            ds.destroy()

        self.__drawQueue.stop()
        self.__drawQueue       = None
        self.__drawnDataSeries = None
        self.dataSeries        = []
        self.artists           = []
        self.__figure          = None
        self.__axis            = None
        self.__destroyed       = True


    @property
    def destroyed(self):
        """Returns True if :meth:`destroy` has been called, ``False``
        otherwise.
        """
        return self.__destroyed


    @property
    def figure(self):
        """Returns the ``matplotlib`` ``Figure`` instance."""
        return self.__figure

    @property
    def axis(self):
        """Returns the ``matplotlib`` ``Axis`` instance."""
        return self.__axis


    @property
    def canvas(self):
        """Returns the ``matplotlib`` ``Canvas`` instance."""
        return self.__canvas


    def getDrawQueue(self):
        """Returns the :class`.idle.TaskThread` instance used for data
        preparation.
        """
        return self.__drawQueue


    def draw(self, *a):
        """Call :meth:`drawDataSeries` and then :meth:`drawArtists`.
        Or, if a ``drawFunc`` was provided, calls that instead.

        You will generally want to call :meth:`asyncDraw` instead of this
        method.
        """
        if self.destroyed:
            return

        if self.__drawFunc:
            self.__drawFunc(*a)
        else:
            self.drawDataSeries()
            self.drawArtists()


    def asyncDraw(self, *a):
        """Schedules :meth:`draw` to be run asynchronously. This method
        should be used in preference to calling :meth:`draw` directly
        in most cases, particularly where the call occurs within a
        property callback function.

        This method is automatically called called whenever a
        :class:`.DataSeries` is added to the :attr:`dataSeries` list, or when
        any plot display properties change.
        """

        # don't run the task if it's
        # already scheduled on the idle loop
        idleName = '{}.draw'.format(id(self))
        if not self.destroyed and not idle.idleLoop.inIdle(idleName):
            idle.idle(self.draw, name=idleName)


    def message(self, msg, clear=True, border=False):
        """Displays the given message in the centre of the figure.

        This is a convenience method provided for use by subclasses.
        """

        axis = self.axis

        if clear:
            self.__drawnDataSeries.clear()
            axis.clear()
            axis.set_xlim((0.0, 1.0))
            axis.set_ylim((0.0, 1.0))

        if border:
            bbox = {'facecolor' : '#ffffff',
                    'edgecolor' : '#cdcdff',
                    'boxstyle'  : 'round,pad=1'}
        else:
            bbox = None

        axis.text(0.5, 0.5,
                  msg,
                  ha='center', va='center',
                  transform=axis.transAxes,
                  bbox=bbox)

        self.canvas.draw()


    def getArtist(self, ds):
        """Returns the ``matplotlib.Artist`` (typically a ``Line2D`` instance)
        associated with the given :class:`.DataSeries` instance. A
        ``KeyError`` is raised if there is no such artist.
        """

        return self.__drawnDataSeries[ds]


    def getDrawnDataSeries(self):
        """Returns a list of tuples, each tuple containing the
        ``(DataSeries, x, y)`` data for one ``DataSeries`` instance
        as it is shown on the plot.
        """
        return [(ds, np.array(l.get_xdata()), np.array(l.get_ydata()))
                for ds, l in self.__drawnDataSeries.items()]


    def prepareDataSeries(self, ds):
        """Prepares the data from the given :class:`.DataSeries` so it is
        ready to be plotted. Called by the :meth:`__drawOneDataSeries` method
        for any ``extraSeries`` passed to the :meth:`drawDataSeries` method
        (but **not** applied to :class:`.DataSeries` that have been added to
        the :attr:`dataSeries` list).

        This implementation just returns :class:`.DataSeries.getData` - you
        can pass a ``prepareFunc`` to ``__init__`` to perform any custom
        preprocessing.
        """
        if self.__prepareFunc:
            return self.__prepareFunc(ds)
        else:
            return ds.getData()


    def drawArtists(self, refresh=True, immediate=False):
        """Draw all ``matplotlib.Artist`` instances in the :attr:`artists`
        list, then refresh the canvas.

        :arg refresh: If ``True`` (default), the canvas is refreshed.
        """

        axis   = self.axis
        canvas = self.canvas

        def realDraw():

            # Just in case this PlotPanel is destroyed
            # before this task gets executed
            if not fwidgets.isalive(self.__canvas):
                return

            for artist in self.artists:
                if artist not in axis.findobj(type(artist)):
                    axis.add_artist(artist)

        if immediate: realDraw()
        else:         self.__drawQueue.enqueue(idle.idle, realDraw)

        def refreshCanvas():
            if not self.destroyed:
                canvas.draw()

        if refresh:
            if immediate: refreshCanvas()
            else:         self.__drawQueue.enqueue(idle.idle, refreshCanvas)


    def drawDataSeries(self, extraSeries=None, refresh=False, **plotArgs):
        """Queues a request to plot all of the :class:`.DataSeries` instances
        in the :attr:`dataSeries` list.

        This method does not do the actual plotting - it is performed
        asynchronously, to avoid locking up the GUI:

         1. The data for each ``DataSeries`` instance is prepared on
            separate threads (using :func:`.idle.run`).

         2. A call to :func:`.idle.wait` is enqueued on a
            :class:`.TaskThread`.

         3. This ``wait`` function waits until all of the data preparation
            threads have completed, and then passes all of the data to
            the :meth:`__drawDataSeries` method.

        :arg extraSeries: A sequence of additional ``DataSeries`` to be
                          plotted. These series are passed through the
                          :meth:`prepareDataSeries` method before being
                          plotted.

        :arg refresh:     If ``True``, the canvas is refreshed. Otherwise,
                          you must call ``getCanvas().draw()`` manually.
                          Defaults to ``False`` - the :meth:`drawArtists`
                          method will refresh the canvas, so if you call
                          :meth:`drawArtists` immediately after calling
                          this method (which you should), then you don't
                          need to manually refresh the canvas.

        :arg plotArgs:    Passed through to the :meth:`__drawDataSeries`
                          method.

        .. note:: This method must only be called from the main application
                  thread (the ``wx`` event loop).
        """

        if extraSeries is None:
            extraSeries = []

        canvas      = self.canvas
        axis        = self.axis
        toPlot      = self.dataSeries[:]

        toPlot      = [ds for ds in toPlot      if ds.enabled]
        extraSeries = [ds for ds in extraSeries if ds.enabled]

        toPlot      = extraSeries + toPlot
        preprocs    = [True] * len(extraSeries) + [False] * len(toPlot)

        if len(toPlot) == 0:
            self.__drawnDataSeries.clear()
            axis.clear()
            canvas.draw()
            return

        # Before clearing/redrawing, save
        # a copy of the x/y axis limits -
        # the user may have changed them
        # via panning/zooming and, if
        # autoLimit is off, we will want
        # to preserve the limits that the
        # user set. These are passed to
        # the __drawDataSeries method.
        #
        # Make sure the limits are ordered
        # as (min, max), as they won't be
        # if invertX/invertY are active.
        axxlim = list(sorted(axis.get_xlim()))
        axylim = list(sorted(axis.get_ylim()))

        # Here we are preparing the data for
        # each data series on separate threads,
        # as data preparation can be time
        # consuming for large images. We
        # display a message on the canvas
        # during preparation.
        tasks    = []
        allXdata = [None] * len(toPlot)
        allYdata = [None] * len(toPlot)

        # Create a separate function
        # for each data series
        for idx, (ds, preproc) in enumerate(zip(toPlot, preprocs)):

            def getData(d=ds, p=preproc, i=idx):

                if not d.enabled:
                    return

                if p: xdata, ydata = self.prepareDataSeries(d)
                else: xdata, ydata = d.getData()

                allXdata[i] = xdata
                allYdata[i] = ydata

            tasks.append(getData)

        # Run the data preparation tasks,
        # a separate thread for each.
        tasks = [idle.run(t) for t in tasks]

        # Show a message while we're
        # preparing the data.
        self.message(strings.messages[self, 'preparingData'],
                     clear=False,
                     border=True)

        # Wait until data preparation is
        # done, then call __drawDataSeries.
        self.__drawRequests += 1
        self.__drawQueue.enqueue(idle.wait,
                                 tasks,
                                 self.__drawDataSeries,
                                 toPlot,
                                 allXdata,
                                 allYdata,
                                 axxlim,
                                 axylim,
                                 refresh,
                                 taskName='{}.wait'.format(id(self)),
                                 wait_direct=True,
                                 **plotArgs)


    def __drawDataSeries(
            self,
            dataSeries,
            allXdata,
            allYdata,
            oldxlim,
            oldylim,
            refresh,
            xlabel=None,
            ylabel=None,
            **plotArgs):
        """Called by :meth:`__drawDataSeries`. Plots all of the data
        associated with the given ``dataSeries``.

        :arg dataSeries: The list of :class:`.DataSeries` instances to plot.

        :arg allXdata:   A list of arrays containing X axis data, one for each
                         ``DataSeries``.

        :arg allYdata:   A list of arrays containing Y axis data, one for each
                         ``DataSeries``.

        :arg oldxlim:    X plot limits from the previous draw. If
                         ``xAutoScale`` is disabled, this limit is preserved.

        :arg oldylim:    Y plot limits from the previous draw. If
                         ``yAutoScale`` is disabled, this limit is preserved.

        :arg refresh:    Refresh the canvas - see :meth:`drawDataSeries`.

        :arg xlabel:     If provided, overrides the value of the :attr:`xlabel`
                         property.

        :arg ylabel:     If provided, overrides the value of the :attr:`ylabel`
                         property.

        :arg plotArgs:   Remaining arguments passed to the
                         :meth:`__drawOneDataSeries` method.
        """

        # Avoid spursious post-destruction
        # notifications (occur sporadically
        # during testing)
        if self.destroyed:
            return

        # Only draw the plot if there are no
        # pending draw requests. Otherwise
        # we would be drawing out-of-date data.
        self.__drawRequests -= 1
        if self.__drawRequests != 0:
            return

        axis          = self.axis
        canvas        = self.canvas
        width, height = canvas.get_width_height()

        self.__drawnDataSeries.clear()
        axis.clear()

        xlims = []
        ylims = []

        for ds, xdata, ydata in zip(dataSeries, allXdata, allYdata):

            if any((ds is None, xdata is None, ydata is None)):
                continue

            if not ds.enabled:
                continue

            xdata      = self.xOffset + self.xScale * xdata
            ydata      = self.yOffset + self.yScale * ydata
            xlim, ylim = self.__drawOneDataSeries(ds,
                                                  xdata,
                                                  ydata,
                                                  **plotArgs)

            if np.any(np.isclose([xlim[0], ylim[0]], [xlim[1], ylim[1]])):
                continue

            xlims.append(xlim)
            ylims.append(ylim)

        if len(xlims) == 0:
            xmin, xmax = 0.0, 0.0
            ymin, ymax = 0.0, 0.0
        else:
            (xmin, xmax), (ymin, ymax) = self.__calcLimits(
                xlims, ylims, oldxlim, oldylim, width, height)

        # x/y axis labels
        if xlabel is None: xlabel = self.xlabel
        if ylabel is None: ylabel = self.ylabel
        if xlabel is None: xlabel = ''
        if ylabel is None: ylabel = ''

        xlabel = xlabel.strip()
        ylabel = ylabel.strip()

        if xlabel != '':
            axis.set_xlabel(xlabel, va='bottom')
            axis.xaxis.set_label_coords(0.5, 10.0 / height)

        if ylabel != '':
            axis.set_ylabel(ylabel, va='top')
            axis.yaxis.set_label_coords(10.0 / width, 0.5)

        # Ticks
        if self.ticks:
            axis.tick_params(direction='in', pad=-5)
            axis.tick_params(axis='both', which='both', length=3)

            for ytl in axis.yaxis.get_ticklabels():
                ytl.set_horizontalalignment('left')

            for xtl in axis.xaxis.get_ticklabels():
                xtl.set_verticalalignment('bottom')
        else:

            # we clear the labels, but
            # leave the ticks, so the
            # axis grid gets drawn
            xlabels = ['' for i in range(len(axis.xaxis.get_ticklabels()))]
            ylabels = ['' for i in range(len(axis.yaxis.get_ticklabels()))]

            axis.set_xticklabels(xlabels)
            axis.set_yticklabels(ylabels)
            axis.tick_params(axis='both', which='both', length=0)

        # Limits
        if xmin != xmax:
            if self.invertX: axis.set_xlim((xmax, xmin))
            else:            axis.set_xlim((xmin, xmax))
            if self.invertY: axis.set_ylim((ymax, ymin))
            else:            axis.set_ylim((ymin, ymax))

        # legend
        labels = [ds.label for ds in dataSeries if ds.label is not None]
        if len(labels) > 0 and self.legend:
            handles, labels = axis.get_legend_handles_labels()
            legend          = axis.legend(
                handles,
                labels,
                loc='upper right',
                fontsize=10,
                handlelength=3,
                fancybox=True)
            legend.get_frame().set_alpha(0.6)

        if self.grid:
            axis.grid(linestyle='-',
                      color=self.gridColour,
                      linewidth=0.5,
                      zorder=0)
        else:
            axis.grid(False)

        axis.spines['right'] .set_visible(False)
        axis.spines['left']  .set_visible(False)
        axis.spines['top']   .set_visible(False)
        axis.spines['bottom'].set_visible(False)

        axis.set_axisbelow(True)
        axis.patch.set_facecolor(self.bgColour)
        self.figure.patch.set_alpha(0)

        if refresh:
            canvas.draw()


    def __drawOneDataSeries(self, ds, xdata, ydata, **plotArgs):
        """Plots a single :class:`.DataSeries` instance. This method is called
        by the :meth:`drawDataSeries` method.

        :arg ds:       The ``DataSeries`` instance.
        :arg xdata:    X axis data.
        :arg ydata:    Y axis data.
        :arg plotArgs: May be used to customise the plot - these
                       arguments are all passed through to the
                       ``Axis.plot`` function.
        """

        if ds.alpha == 0:
            return (0, 0), (0, 0)

        if len(xdata) != len(ydata) or len(xdata) == 0:
            log.debug('{}: data series length mismatch, or '
                      'no data points (x: {}, y: {})'.format(
                          ds.overlay.name, len(xdata), len(ydata)))
            return (0, 0), (0, 0)

        xdata = np.asarray(xdata, dtype=float)
        ydata = np.asarray(ydata, dtype=float)

        log.debug('Drawing {} for {}'.format(type(ds).__name__, ds.overlay))

        # Note to self: If the smoothed data is
        # filled with NaNs, it is possibly due
        # to duplicate values in the x data, which
        # are not handled very well by splrep.
        if self.smooth:

            tck   = interp.splrep(xdata, ydata)
            xdata = np.linspace(xdata[0],
                                xdata[-1],
                                len(xdata) * 5,
                                dtype=np.float32)
            ydata = interp.splev(xdata, tck)

        nans        = ~(np.isfinite(xdata) & np.isfinite(ydata))
        xdata[nans] = np.nan
        ydata[nans] = np.nan

        if self.xLogScale: xdata[xdata <= 0] = np.nan
        if self.yLogScale: ydata[ydata <= 0] = np.nan

        if np.all(np.isnan(xdata) | np.isnan(ydata)):
            return (0, 0), (0, 0)

        kwargs = plotArgs

        kwargs['lw']    = kwargs.get('lw',    ds.lineWidth)
        kwargs['alpha'] = kwargs.get('alpha', ds.alpha)
        kwargs['color'] = kwargs.get('color', ds.colour)
        kwargs['label'] = kwargs.get('label', ds.label)
        kwargs['ls']    = kwargs.get('ls',    ds.lineStyle)

        axis = self.axis
        line = axis.plot(xdata, ydata, **kwargs)[0]

        self.__drawnDataSeries[ds] = line

        if self.xLogScale:
            axis.set_xscale('log')
            posx    = xdata[xdata > 0]
            xlimits = np.nanmin(posx), np.nanmax(posx)

        else:
            xlimits = np.nanmin(xdata), np.nanmax(xdata)

        if self.yLogScale:
            axis.set_yscale('log')
            posy    = ydata[ydata > 0]
            ylimits = np.nanmin(posy), np.nanmax(posy)
        else:
            ylimits = np.nanmin(ydata), np.nanmax(ydata)

        return xlimits, ylimits


    def __dataSeriesChanged(self, *a):
        """Called when the :attr:`dataSeries` list changes. Adds listeners
        to any new :class:`.DataSeries` instances, and then calls
        :meth:`asyncDraw`.
        """

        for ds in self.dataSeries:
            for propName in ds.redrawProperties():
                ds.addListener(propName,
                               self.__name,
                               self.asyncDraw,
                               overwrite=True)
        self.asyncDraw()


    def __artistsChanged(self, *a):
        """Called when the :attr:`artists` list changes. Calls
        :meth:`asyncDraw`.
        """
        self.asyncDraw()


    def __limitsChanged(self, *a):
        """Called when the :attr:`limits` change. Updates the axis limits
        accordingly.
        """

        axis = self.axis
        axis.set_xlim(self.limits.x)
        axis.set_ylim(self.limits.y)
        self.asyncDraw()


    def __calcLimits(self,
                     dataxlims,
                     dataylims,
                     axisxlims,
                     axisylims,
                     axWidth,
                     axHeight):
        """Calculates and returns suitable axis limits for the current plot.
        Also updates the :attr:`limits` property. This method is called by
        the :meth:`drawDataSeries` method.

        If :attr:`xAutoScale` or :attr:`yAutoScale` are enabled, the limits are
        calculated from the data range, using the canvas width and height to
        maintain consistent padding around the plotted data, irrespective of
        the canvas size.

        . Otherwise, the existing axis limits are retained.

        :arg dataxlims: A tuple containing the (min, max) x data range.

        :arg dataylims: A tuple containing the (min, max) y data range.

        :arg axisxlims: A tuple containing the current (min, max) x axis
                        limits.

        :arg axisylims: A tuple containing the current (min, max) y axis
                        limits.

        :arg axWidth:   Canvas width in pixels

        :arg axHeight:  Canvas height in pixels
        """

        if self.xAutoScale:

            xmin = min([lim[0] for lim in dataxlims])
            xmax = max([lim[1] for lim in dataxlims])

            lPad = (xmax - xmin) * (50.0 / axWidth)
            rPad = (xmax - xmin) * (50.0 / axWidth)

            xmin = xmin - lPad
            xmax = xmax + rPad
        else:
            xmin = axisxlims[0]
            xmax = axisxlims[1]

        if self.yAutoScale:

            ymin = min([lim[0] for lim in dataylims])
            ymax = max([lim[1] for lim in dataylims])

            bPad = (ymax - ymin) * (50.0 / axHeight)
            tPad = (ymax - ymin) * (50.0 / axHeight)

            ymin = ymin - bPad
            ymax = ymax + tPad

        else:

            ymin = axisylims[0]
            ymax = axisylims[1]

        self.disableListener('limits', self.__name)
        self.limits[:] = [xmin, xmax, ymin, ymax]
        self.enableListener('limits', self.__name)

        return (xmin, xmax), (ymin, ymax)



#!/usr/bin/env python
#
# viewpanel.py - Superclass for all FSLeyes view panels.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ViewPanel` class, which is the base-class
for all of the *FSLeyes view* panels. See the :mod:`fsleyes` package
documentation for more details.
"""


import              logging
import itertools as it

import                                  wx
import wx.lib.agw.aui                as aui
import wx.lib.agw.aui.framemanager   as auifm

import fsl.utils.notifier            as notifier
import fsleyes_widgets               as fwidgets

import fsleyes.panel                 as fslpanel
import fsleyes.toolbar               as fsltoolbar
import fsleyes.plugins               as plugins
import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.profiles              as profiles
import fsleyes.strings               as strings
import fsleyes.actions               as actions


log = logging.getLogger(__name__)


class ViewPanel(fslpanel.FSLeyesPanel):
    """The ``ViewPanel`` class is the base-class for all *FSLeyes views*.

    A ``ViewPanel`` displays some sort of view of the overlays in an
    :class:`.OverlayList`. The settings for a ``ViewPanel`` are defined by a
    :class:`.DisplayContext` instance.


    **Panels and controls**


    A ``ViewPanel`` class uses a ``wx.lib.agw.aui.AuiManager`` to lay out its
    children. A ``ViewPanel`` has one central panel, which contains the
    primary view; and may have one or more secondary panels, which contain
    *controls* - see the :mod:`.controls` package. The centre panel can be set
    via the :meth:`centrePanel` property, and secondary panels can be
    added/removed to/from with the :meth:`togglePanel` method. The current
    state of a secondary panel (i.e. whether one is open or not) can be
    queried with the :meth:`isPanelOpen` method, and existing secondary panels
    can be accessed via the :meth:`getPanel` method.  Secondary panels must be
    derived from either the :class:`.ControlPanel` or :class:`.ControlToolBar`
    base-classes.


    **Profiles**


    The logic which defines the way that a user interacts with a view panel
    is defined by a  :class:`.Profile` object, which contains mouse and
    keyboard event handlers for reacting to user input.

    Sub-classes must call the :meth:`initProfile` method to initialise their
    default profile. Other profiles may be associated with a particular
    control panel - these profiles can be temporarily activated and deactivated
    when the control panel is toggled via the :meth:`togglePanel` method.

    The currently active interaction profile can be accessed with the
    :meth:`currentProfile` method. See the :mod:`.profiles` package for
    more information on interaction profiles.


    **Programming interface**


    The following methods are available on a ``Viewpanel`` for programmatically
    controlling its display and layout:


    .. autosummary::
       :nosignatures:

       togglePanel
       isPanelOpen
       getPanel
       getPanels
       getTools
       removeFromFrame
       removeAllPanels
       getPanelInfo
       auiManager
    """


    def controlOptions(self, cpType):
        """May be overridden by sub-classes. Given a control panel type,
        may return a dictionary containing arguments to be passed to
        the  ``__init__`` method when the control panel is created.
        """
        return None


    @staticmethod
    def controlOrder():
        """May be overridden by sub-classes. Returns a list of names of
        control panel types, specifying a suggested order for the
        settings menu for views of this type.
        """
        return None


    @staticmethod
    def toolOrder():
        """May be overridden by sub-classes. Returns a list of names of
        tools, specifying a suggested order for the corresponding entries
        in the FSLeyes tools menu. Note that the ordering of tools returned
        by the :meth:`getTools` method is honoured - the ordering returned
        by *this* method relates to tools which are implemented as plugins.
        """
        return None


    @staticmethod
    def defaultLayout():
        """May be overridden by sub-classes. Should return a list of names of
        FSLeyes :class:`.ControlPanel` types which form the default layout for
        this view.
        """
        return None


    def __init__(self, parent, overlayList, displayCtx, frame):
        """Create a ``ViewPanel``.

        :arg parent:         ``wx`` parent object
        :arg overlayList:    The :class:`.OverlayList`
        :arg displayCtx:     A :class:`.DisplayContext` object unique to this
                             ``ViewPanel``
        :arg frame:          The :class:`.FSLeyesFrame`
        """

        fslpanel.FSLeyesPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        # Currently only two profiles are allowed at
        # any one time - the default profile, passed
        # to initProfiles. and one more profile
        # associated with a control panel.
        #
        # If the default profile is active, and a
        # new control is opened which needs a different
        # profile, that profile is activated.
        #
        # If a non-default profile (#1) is active, and
        # a new control is opened which needs a
        # different profile (#2), all controls which
        # require profile #1 are closed, and profile
        # #2 is activated.
        #
        # This all happens in togglePanel.
        self.__profileManager = profiles.ProfileManager(
            self, overlayList, displayCtx, 2)

        # The __centrePanel attribute stores a reference
        # to the main (centre) panel on this ViewPanel.
        # It is set by sub-class implementations via
        # the centrePanel property.
        #
        # The panels dictionary stores a collection
        # of {type : instance} mappings of active
        # FSLeyes control panels that are contained
        # in this view panel.
        self.__centrePanel = None
        self.__panels      = {}

        # Notifier instance for emitting events.
        # Currently only two events are emitted
        # - one of them is profile changes, and
        # the profilemanager emits these events
        # anyway, so we can just use it. As for
        # the other event (aui layout changes),
        # well, we're being a bit dodgy and
        # emitting these events via the
        # profilemanager.
        self.__events = self.__profileManager

        # See note in FSLeyesFrame about
        # the user of aero docking guides.
        self.__auiMgr = aui.AuiManager(
            self,
            agwFlags=(aui.AUI_MGR_RECTANGLE_HINT          |
                      aui.AUI_MGR_NO_VENETIAN_BLINDS_FADE |
                      aui.AUI_MGR_ALLOW_FLOATING          |
                      aui.AUI_MGR_AERO_DOCKING_GUIDES     |
                      aui.AUI_MGR_LIVE_RESIZE))

        self.__auiMgr.SetDockSizeConstraint(0.5, 0.5)
        self.__auiMgr.Bind(aui.EVT_AUI_PANE_CLOSE,
                           self.__onPaneClose)
        self.__auiMgr.Bind(aui.EVT_AUI_PERSPECTIVE_CHANGED,
                           self.__onPerspectiveChange)

        # A very shitty necessity. When panes are floated,
        # the AuiManager sets the size of the floating frame
        # to the minimum size of the panel, without taking
        # into account the size of its borders/title bar,
        # meaning that the panel size is too small. Here,
        # we're just creating a dummy MiniFrame (from which
        # the AuiFloatingFrame derives), and saving the size
        # of its trimmings for later use in the togglePanel
        # method.
        ff         = wx.MiniFrame(self)
        size       = ff.GetSize().Get()
        clientSize = ff.GetClientSize().Get()
        self.__floatOffset = (size[0] - clientSize[0],
                              size[1] - clientSize[1])
        ff.Destroy()

        self.__loadPlugins()


    def __loadPlugins(self):
        """Called by :meth:`__init__`. This is a bit of a hack, but less hacky
        than it used to be.  All plugin-provided control panels and tools
        which support this ``ViewPanel`` are looked up via the :mod:`.plugins.`
        module. Then, for each control, we create a
        :class:`.ToggleControlPanelAction`, and add it as an attribute on this
        ``ViewPanel``.

        Similarly, all plugin-provided tools which support this ``ViewPanel``
        are created and added as attributes.

        In both cases, the class name of the control/tool is used as the
        attribute name.

        This is done so that these actions will work with the
        :class:`.ActionProvider` interface, and hence the
        :meth:`.FSLEeyesFrame.populateMenu` method.

        This implementation may change in the future if it becomes problematic
        (e.g. due to naming conflicts).
        """

        # controls
        for ctrlType in plugins.listControls(type(self)).values():
            name = ctrlType.__name__
            act  = actions.ToggleControlPanelAction(
                self.overlayList,
                self.displayCtx,
                ctrlType,
                self,
                name=name)
            setattr(self, name, act)

        # tools
        for toolType in plugins.listTools(type(self)).values():
            name = toolType.__name__
            act  = toolType(self.overlayList, self.displayCtx, self)
            setattr(self, name, act)


    def destroy(self):
        """Removes some property listeners, destroys all child panels,
        destroys the :class:`.ProfileManager`, and ``AuiManager``, and
        calls :meth:`.FSLeyesPanel.destroy`.
        """

        # Order of operations is important. For example,
        # FSLeyesPanel.destroy will result in all Actions
        # associated with this ViewPanel being destroyed,
        # and their destroy routines may interact with
        # this ViewPanel.

        # remove listeners from overlaylist and display context
        lName = 'ViewPanel_{}'.format(self.name)
        self.overlayList.removeListener('overlays',        lName)
        self.displayCtx .removeListener('selectedOverlay', lName)

        fslpanel.FSLeyesPanel.destroy(self)

        # Make sure that any control panels are correctly destroyed
        for panel in self.__panels.values():
            self.__auiMgr.DetachPane(panel)
            panel.destroy()

        # Disable the  ProfileManager
        self.__profileManager.destroy()

        # Un-initialise the AUI manager
        self.__auiMgr.Unbind(aui.EVT_AUI_PANE_CLOSE)
        self.__auiMgr.Unbind(aui.EVT_AUI_PERSPECTIVE_CHANGED)
        self.__auiMgr.Update()
        self.__auiMgr.UnInit()

        # The AUI manager does not clear its
        # reference to this panel, so let's
        # do it here.
        self.__auiMgr._frame  = None
        self.__profileManager = None
        self.__auiMgr         = None
        self.__panels         = None
        self.__centrePanel    = None
        self.__events         = None


    @property
    def events(self) -> notifier.Notifier:
        """Return a reference to a :class:`.Notifier` instance which can be
        used to be notified when certain events occur. Currently the following
        events are emitted:

         - ``'profile'``, when the current interaction profile changes.
           Callbacks which are registered with the ``'profile'`` topic will
           be passed a tuple containing the types (:class:`.Profile`
           sub-classes) of the de-registered and newly registered profiles.

         - ``'aui_perspective'``, when the AUI-managed layout changes, e.g.
           sash resizes, control panels added/removed, etc. This event is
           emitted whenever the ``AuiManager`` emits a
           ``EVT_AUI_PERSPECTIVE_CHANGED`` event. It is re-emitted via the
           :class:`.Notifer` interface so that non-wx entities can be
           notified (see e.g. the :class:`.ToggleControlPanelAction`).
        """
        return self.__events


    def initProfile(self, defaultProfile):
        """Must be called by subclasses, after they have initialised all
        of the attributes which may be needed by their associated
        :class:`.Profile` instances.

        :arg defaultProfile: Default profile type
        """
        self.__profileManager.activateProfile(defaultProfile)


    @property
    def currentProfile(self):
        """Returns the :class:`.Profile` instance currently in use. """
        return self.__profileManager.getCurrentProfile()


    @property
    def profileManager(self):
        """Returns a reference to the :class:`.ProfileManager` used by
        this ``ViewPanel``.
        """
        return self.__profileManager


    @property
    def centrePanel(self):
        """Returns the primary (centre) panel on this ``ViewPanel``.
        """
        return self.__centrePanel


    @centrePanel.setter
    def centrePanel(self, panel):
        """Set the primary centre panel for this ``ViewPanel``. This method
        is only intended to be called by sub-classes.
        """

        panel.Reparent(self)
        paneInfo = (aui.AuiPaneInfo()
                    .Name(type(panel).__name__)
                    .CentrePane())

        self.__auiMgr.AddPane(panel, paneInfo)
        self.__auiMgrUpdate()
        self.__centrePanel = panel


    def togglePanel(self, panelType, *args, **kwargs):
        """Add/remove the secondary control panel of the specified type to/from
        this ``ViewPanel``.

        If no keyword argunents are specified, the arguments returned by the
        :meth:`.ControlMixin.defaultLayout` method are returned.

        :arg panelType: Type of the secondary panel.

        :arg args:      All positional arguments are passed to the
                        ``panelType`` constructor.

        :arg floatPane: If ``True``, the secondary panel is initially floated.
                        Defaults to ``False``.

        :arg floatOnly: If ``True``, and ``floatPane=True``, the panel will
                        be permanently floated (i.e. it will not be dockable).

        :arg floatPos:  If provided, and ``floatPane`` is ``True``, specifies
                        the location of the floating panel as ``(w, h)``
                        proportions between 0 and 1, relative to this view
                        panel.

        :arg closeable: If ``False``, and ``floatPane=True``, the panel will
                        not have a close button when it is floated.

        :arg location:  If ``floatPane=False``, the initial dock position of
                        the panel - either ``wx.TOP``, ``wx.BOTTOM``,
                        ``wx.LEFT``, or ``wx.RIGHT. Defaults to ``wx.BOTTOM``.

        :arg title:     Title to give the control. If not provided, it is
                        assumed that a title for ``panelType`` is in
                        :attr:`.strings.titles`.

        :arg kwargs:    All other keyword arguments, are passed to the
                        ``panelType`` constructor.

        :returns:       The newly created control panel.

        .. note::       The ``panelType`` type must be a sub-class of
                        :class:`.ControlPanel` or :class:`.ControlToolBar`,
                        which can be created like so::

                            panel = panelType(parent,
                                              overlayList,
                                              displayCtx,
                                              frame,
                                              *args,
                                              **kwargs)

        .. warning:: Do not define a control (a.k.a. secondary) panel
                     constructor to accept arguments with the names
                     ``floatPane``, ``floatOnly``, ``floatPos``,
                     ``closeable``, or ``location``, as arguments with those
                     names will get eaten by this method before they can be
                     passed to the constructor.
        """

        # Merge in default options for this control,
        # with precedence (highest to lowest):
        #  1. kwargs
        #  2. ViewPanel.controlOptions
        #  3. ControlPanel.defaultLayout
        layout = panelType.defaultLayout()
        cpopts = self.controlOptions(panelType)

        if layout is None: layout = {}
        if cpopts is None: cpopts = {}

        for k, v in it.chain(cpopts.items(), layout.items()):
            if k not in kwargs:
                kwargs[k] = v

        location   = kwargs.pop('location',  None)
        floatPane  = kwargs.pop('floatPane', False)
        floatOnly  = kwargs.pop('floatOnly', False)
        closeable  = kwargs.pop('closeable', True)
        title      = kwargs.pop('title',     None)
        floatPos   = kwargs.pop('floatPos',  (0.5, 0.5))
        profileCls = panelType.profileCls()

        if title is None:
            title = strings.titles.get(panelType, type(panelType).__name__)

        if location not in (None, wx.TOP, wx.BOTTOM, wx.LEFT, wx.RIGHT):
            raise ValueError('Invalid value for location')

        supported = panelType.supportedViews()
        if supported is not None and not isinstance(self, tuple(supported)):
            raise ValueError(
                '{} views are not supported by {} controls'.format(
                    type(self).__name__, panelType.__name__))

        # The panel is already open - close it
        window = self.__panels.get(panelType, None)
        if window is not None:
            self.__onPaneClose(None, window)
            return

        # Otherwise, create a new panel of the specified type.
        # If this control is associated with a custom interaction
        # profile, check that that profile is active, create it
        # if needed, and close down any controls which are not
        # compatible with the new profile.
        profile = self.currentProfile
        if profileCls is not None and not isinstance(profile, profileCls):

            log.debug('New control %s requires interaction profile %s but '
                      'profile %s is active', panelType.__name__,
                      profileCls.__name__, type(profile).__name__)

            # close down incompatible controls
            for ctype, cpanel in list(self.__panels.items()):
                cprofileCls = ctype.profileCls()
                if cprofileCls is not None and \
                   not issubclass(cprofileCls, profileCls):
                    log.debug('Closing down control %s as it is incompatible '
                              'with profile %s', ctype.__name__,
                              profileCls.__name__)
                    self.__onPaneClose(None, cpanel)

            # change profile
            self.__profileManager.activateProfile(profileCls)

        # The PaneInfo Name is the control panel class name -
        # this is used for saving and restoring layouts.
        window = panelType(self,
                           self.overlayList,
                           self.displayCtx,
                           self,
                           *args,
                           **kwargs)
        # save a ref to the control if it has a custom
        # profile so, when it is destroyed, we know
        if profileCls is not None:
            self.__activeControl = window

        paneInfo = self.__layoutNewPanel(window, title, location, floatPane,
                                         floatPos, floatOnly, closeable)
        self.__auiMgr.AddPane(window, paneInfo)
        self.__panels[panelType] = window
        self.__auiMgrUpdate(newPanel=window)
        return window


    def __layoutNewPanel(self,
                         panel,
                         title,
                         location,
                         floatPane,
                         floatPos,
                         floatOnly,
                         closeable):
        """Sub-method of :meth:`togglePanel`. Creates and returns an
        ``AuiPaneInfo`` instance describing the initial layout of a new
        control panel. See ``togglePanel`` for an explanation of the
        arguments.
        """
        paneInfo  = aui.AuiPaneInfo().Name(type(panel).__name__)
        isToolbar = isinstance(panel, ctrlpanel.ControlToolBar)

        if isToolbar:

            # ToolbarPane sets the panel layer to 10
            paneInfo.ToolbarPane()

            if panel.GetOrient() == wx.VERTICAL:
                paneInfo.GripperTop()

            # We are going to put any new toolbars on
            # the top of the panel, below any existing
            # toolbars. This is annoyingly complicated,
            # because the AUI designer(s) decided to
            # give the innermost layer an index of 0.
            #
            # So in order to put a new toolbar at the
            # innermost layer, we need to adjust the
            # layers of all other existing toolbars

            for p in self.__panels.values():
                if isinstance(p, ctrlpanel.ControlToolBar):
                    info = self.__auiMgr.GetPane(p)

                    # This is nasty - the agw.aui.AuiPaneInfo
                    # class doesn't have any publicly documented
                    # methods of querying its current state.
                    # So I'm accessing its undocumented instance
                    # attributes (determined by browsing the
                    # source code)
                    if info.IsDocked() and \
                       info.dock_direction == aui.AUI_DOCK_TOP:
                        info.Layer(info.dock_layer + 1)

            # When the toolbar contents change,
            # update the layout, so that the
            # toolbar's new size is accommodated
            panel.Bind(fsltoolbar.EVT_TOOLBAR_EVENT, self.__auiMgrUpdate)

        paneInfo.Caption(title)

        # Dock the pane at the position specified
        # by the location parameter
        if not floatPane:

            if location is None:
                if isToolbar: location = aui.AUI_DOCK_TOP
                else:         location = aui.AUI_DOCK_BOTTOM

            elif location == wx.TOP:    location = aui.AUI_DOCK_TOP
            elif location == wx.BOTTOM: location = aui.AUI_DOCK_BOTTOM
            elif location == wx.LEFT:   location = aui.AUI_DOCK_LEFT
            elif location == wx.RIGHT:  location = aui.AUI_DOCK_RIGHT

            # Make sure the pane is
            # resizable in case it
            # gets floated later on
            paneInfo.Direction(location) \
                    .Resizable(not isToolbar)

        # Or, for floating panes, centre the
        # floating pane on this ViewPanel
        else:

            selfPos    = self.GetScreenPosition().Get()
            selfSize   = self.GetSize().Get()
            selfCentre = (selfPos[0] + selfSize[0] * floatPos[0],
                          selfPos[1] + selfSize[1] * floatPos[1])

            paneSize = panel.GetBestSize().Get()
            panePos  = (selfCentre[0] - paneSize[0] * 0.5,
                        selfCentre[1] - paneSize[1] * 0.5)

            paneInfo.Float()                 \
                    .Resizable(True)         \
                    .Dockable(not floatOnly) \
                    .CloseButton(closeable)  \
                    .FloatingPosition(panePos)

        return paneInfo


    def isPanelOpen(self, panelType):
        """Returns ``True`` if a panel of type ``panelType`` is open,
        ``False`` otherwise.
        """
        return self.getPanel(panelType) is not None


    def getPanel(self, panelType):
        """If an instance of ``panelType`` exists, it is returned.
        Otherwise ``None`` is returned.
        """
        if panelType in self.__panels: return self.__panels[panelType]
        else:                          return None


    @actions.action
    def removeAllPanels(self):
        """Remove all control panels from this ``ViewPanel``."""
        for panelType in list(self.__panels.keys()):
            self.togglePanel(panelType)


    @actions.action
    def removeFromFrame(self):
        """Remove this ``ViewPanel`` from the :class:`.FSLeyesFrame`.

        Will raise an error if this ``ViewPanel`` is not in a
        ``FSLeyesFrame``.
        """
        self.frame.removeViewPanel(self)


    def getPanels(self):
        """Returns a list containing all control panels currently shown in this
        ``ViewPanel``.
        """
        return list(self.__panels.values())


    def getPanelInfo(self, panel):
        """Returns the ``AuiPaneInfo`` object which contains information about
        the given control panel.
        """
        return self.__auiMgr.GetPane(panel)


    @property
    def auiManager(self):
        """Returns the ``wx.lib.agw.aui.AuiManager`` object which manages the
        layout of this ``ViewPanel``.
        """
        return self.__auiMgr


    def getTools(self):
        """This method should be overridden by sub-classes (if necessary), and
        should return any ``action`` methods which should be added to the
        :class:`.FSLeyesFrame` *Tools* menu.

        See also the :meth:`.ActionProvider.getActions` method, which can
        also be overridden, and controls the actions which get added to the
        FSLeyes *settings* menu.
        """
        return []


    def __auiMgrUpdate(self, *args, **kwargs):
        """Called whenever a panel is added/removed to/from this ``ViewPanel``.

        Calls the ``Update`` method on the ``AuiManager`` instance that is
        managing this panel.

        :arg newPanel: Must be passed as a keyword argument. When a new panel
                       is added, it should be passed here.
        """

        newPanel = kwargs.pop('newPanel', None)

        # This method makes sure that size hints
        # for all existing and new panels are
        # set on their AuiPaneInfo objects, and
        # then calls AuiManager.Update.

        # We first loop through all panels, and
        # figure out their best sizes. Each entry
        # in this list is a tuple containing:
        #
        #    - Panel
        #    - AuiPaneInfo instance
        #    - Dock direction (None for floating panels)
        #    - Layer number (None for floating panels)
        #    - Minimum size
        bestSizes = []

        for panel in self.__panels.values():

            if isinstance(panel, ctrlpanel.ControlToolBar):
                continue

            pinfo = self.__auiMgr.GetPane(panel)

            # If the panel is floating, use its
            # current size as its 'best' size,
            # as otherwise the AuiManager will
            # immediately resize the panel to
            # its best size.
            if pinfo.IsFloating():
                dockDir  = None
                layer    = None
                bestSize = panel.GetSize().Get()

                # Unless its current size is tiny
                # (which probably means that it has
                # just been added)
                if bestSize[0] <= 20 or \
                   bestSize[1] <= 20:
                    bestSize = panel.GetBestSize().Get()

            else:
                dockDir  = pinfo.dock_direction
                layer    = pinfo.dock_layer
                bestSize = panel.GetBestSize().Get()

            bestSizes.append((panel, pinfo, dockDir, layer, bestSize))

        # Now we loop through one final time, and
        # set all of the necessary size hints on
        # the AuiPaneInfo instances.
        for panel, pinfo, dockDir, layer, bestSize in bestSizes:

            parent = panel.GetParent()

            # When a panel is added/removed from the AuiManager,
            # the position of floating panels seems to get reset
            # to their original position, when they were created.
            # Here, we explicitly set the position of each
            # floating frame, so the AuiManager doesn't move our
            # windows about the place.
            if pinfo.IsFloating() and \
               isinstance(parent, aui.AuiFloatingFrame):
                pinfo.FloatingPosition(parent.GetScreenPosition())

            # See comments in __init__ about
            # this silly 'float offset' thing
            floatSize = (bestSize[0] + self.__floatOffset[0],
                         bestSize[1] + self.__floatOffset[1])

            log.debug('New size for panel {} - '
                      'best: {}, float: {}'.format(
                          type(panel).__name__, bestSize, floatSize))

            pinfo.MinSize(     (1, 1))  \
                 .BestSize(    bestSize) \
                 .FloatingSize(floatSize) \
                 .Resizable(   True)

            # This is a terrible hack which forces
            # the AuiManager to grow a dock when a
            # new panel is added, which is bigger
            # than the existing dock contents.
            if panel is newPanel and not pinfo.IsFloating():
                docks = aui.FindDocks(self.__auiMgr._docks, dockDir, layer)
                for d in docks:
                    d.size = 0

        self.__auiMgr.Update()


    def __onPaneClose(self, ev=None, panel=None):
        """Called when the user closes a control (a.k.a. secondary) panel.
        Calls the
        :class:`.ControlPanel.destroy`/:class:`.ControlToolBar.destroy`
        method on the panel.
        """

        if ev is not None:
            ev.Skip()
            panel = ev.GetPane().window

        # If the user has grouped multiple control panels
        # into a single tabbed notebook, and then closed
        # the entire notebook, the AuiManager will generate
        # a single close event, and will pass us that
        # notebook. So we have to look in the notebook
        # to see which control panels were actually closed.
        if isinstance(panel, wx.lib.agw.aui.AuiNotebook):
            panels = [panel.GetPage(i) for i in range(panel.GetPageCount())]
        else:
            panels = [panel]

        for panel in list(panels):

            # note: in theory, all panels  should be sub-classes
            # of ControlPanel/ControlToolBar. But this check is
            # kept here to support third party scripts which don't
            # honour the FSLeyes plugin rules.
            if isinstance(panel, (ctrlpanel.ControlPanel,
                                  ctrlpanel.ControlToolBar)):

                # WTF AUI. Sometimes this method gets called
                # twice for a panel, the second time with a
                # reference to a wx._wxpyDeadObject; in such
                # situations, the Destroy method call below
                # would result in an exception being raised.
                if self.__panels.pop(type(panel), None) is None:
                    panels.remove(panel)

                # calling ControlPanel.destroy()
                # here -  wx.Destroy is done below
                else:
                    log.debug('Panel closed: %s', type(panel).__name__)
                    panel.destroy()

        # Destroy all the panels
        for panel in panels:

            # Even when the user closes a pane,
            # AUI does not detach said pane -
            # we have to do it manually
            self.__auiMgr.DetachPane(panel)
            wx.CallAfter(panel.Destroy)

        # Update interaction profile. We do not
        # consider multiple tabbed controls here.
        # If the closed control is associated with
        # an interaction profile, if no other other
        # open panels rely on the same destroy the
        # profile and restore the default profile
        #
        # We assume that, if a panel which requires
        # a custom profile was open, that profile
        # was active.
        #
        # See WTF AUI comment above for reason for
        # len(panels) guard
        if len(panels) > 0:
            profileCls   = panels[0].profileCls()
            closeProfile = True
            if profileCls is not None:
                for ctype in self.__panels:
                    cprofileCls = ctype.profileCls()
                    if cprofileCls is not None and \
                       issubclass(cprofileCls, profileCls):
                        closeProfile = False
                if closeProfile:
                    log.debug('Panel %s uses a custom interaction profile %s - '
                              'deactivating it and restoring default profile',
                              type(panels[0]).__name__, profileCls.__name__)
                    self.__profileManager.deactivateProfile()

        # Update the view panel layout
        wx.CallAfter(self.__auiMgrUpdate)


    def __onPerspectiveChange(self, ev):
        """Called on ``EVT_AUI_PERSPECTIVE_CHANGED`` events. Re-emits the
        event via the :meth:`events` notifier, with topic ``'aui_perspective'``.
        This is performed for the benefit of non-wx entities which need to
        know about layout changes.
        """
        self.events.notify(topic='aui_perspective')


class MyAuiFloatingFrame(auifm.AuiFloatingFrame):
    """Here I am monkey patching the
    ``wx.agw.aui.framemanager.AuiFloatingFrame.__init__`` method.

    I am doing this because I have observed some strange behaviour when running
    a remote instance of this application over an SSH/X11 session, with the X11
    server (i.e. the local machine) running in OS X. When a combobox is embedded
    in a floating frame (either a pane or a toolbar), its dropdown list appears
    underneath the frame, meaning that the user is unable to actually select any
    items from the list!

    I have only seen this behaviour when using XQuartz on macOS.

    Ultimately, this appears to be caused by the ``wx.FRAME_TOOL_WINDOW``
    style, as passed to the ``wx.MiniFrame`` constructor (from which the
    ``AuiFloatingFrame`` class derives). Removing this style flag fixes the
    problem, so this is exactly what I'm doing. I haven't looked any deeper
    into the situation.


    This class also overrieds the ``SetPaneWindow`` method, because under gtk3,
    the maximum size if a frame musr be set.
    """

    def __init__(self, *args, **kwargs):
        """My new constructor, which makes sure that the ``FRAME_TOOL_WINDOW``
        style is not passed through to the ``AuiFloatingFrame`` constructor
        """

        if 'style' in kwargs:
            style = kwargs['style']

        # This is the default style, as defined
        # in the AuiFloatingFrame constructor
        else:
            style = (wx.FRAME_TOOL_WINDOW     |
                     wx.FRAME_FLOAT_ON_PARENT |
                     wx.FRAME_NO_TASKBAR      |
                     wx.CLIP_CHILDREN)

        if fwidgets.inSSHSession():
            style &= ~wx.FRAME_TOOL_WINDOW

        kwargs['style'] = style

        super().__init__(*args, **kwargs)


    def SetPaneWindow(self, pane):
        """Make sure that floated toolbars are sized correctly.
        """
        super().SetPaneWindow(pane)
        if isinstance(pane.window, ctrlpanel.ControlToolBar):
            size = self.GetBestSize()
            self.SetMaxSize(size)


def _AuiDockingGuide_init(self, *args, **kwargs):
    """I am also monkey-patching the
    ``wx.lib.agw.aui.AuiDockingGuide.__init__`` method, because in this
    instance, when running over SSH/X11, the ``wx.FRAME_TOOL_WINDOW`` style
    seems to result in the docking guide frames being given title bars, which
    is quite undesirable.

    I cannot patch the entire class in the aui package, because it is used
    as part of a class hierarchy. So I am just patching the method.
    """

    if 'style' in kwargs:
        style = kwargs['style']

    # This is the default style, as defined
    # in the AuiDockingGuide constructor
    else:
        style = (wx.FRAME_TOOL_WINDOW |
                 wx.FRAME_STAY_ON_TOP |
                 wx.FRAME_NO_TASKBAR  |
                 wx.NO_BORDER)

    if fwidgets.inSSHSession():
        style &= ~wx.FRAME_TOOL_WINDOW

    kwargs['style'] = style

    _AuiDockingGuide_real_init(self, *args, **kwargs)


aui  .AuiFloatingFrame       = MyAuiFloatingFrame
auifm.AuiFloatingFrame       = MyAuiFloatingFrame
_AuiDockingGuide_real_init   = aui.AuiDockingGuide.__init__
aui.AuiDockingGuide.__init__ = _AuiDockingGuide_init



#!/usr/bin/env python
#
# data.py - Functions for preparing OpenGL texture data.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functions for working with OpenGL texture data:

.. autosummary::
   :nosignatures:

   numTextureDims
   canUseFloatTextures
   oneChannelFormat
   getTextureType
   prepareData
"""


import logging
import inspect

import numpy                       as np
import OpenGL.GL                   as gl
import OpenGL.extensions           as glexts
import OpenGL.GL.ARB.texture_float as arbtf

import fsl.utils.memoize           as memoize
import fsl.transform.affine        as affine
import fsleyes.gl.routines         as glroutines


log = logging.getLogger(__name__)


# Used for log messages
GL_TYPE_NAMES = {

    gl.GL_UNSIGNED_BYTE       : 'GL_UNSIGNED_BYTE',
    gl.GL_UNSIGNED_SHORT      : 'GL_UNSIGNED_SHORT',
    gl.GL_FLOAT               : 'GL_FLOAT',

    gl.GL_ALPHA               : 'GL_ALPHA',
    gl.GL_RED                 : 'GL_RED',
    gl.GL_LUMINANCE           : 'GL_LUMINANCE',
    gl.GL_RGB                 : 'GL_RGB',
    gl.GL_RGBA                : 'GL_RGBA',

    gl.GL_LUMINANCE8          : 'GL_LUMINANCE8',
    gl.GL_LUMINANCE16         : 'GL_LUMINANCE16',
    arbtf.GL_LUMINANCE16F_ARB : 'GL_LUMINANCE16F',
    arbtf.GL_LUMINANCE32F_ARB : 'GL_LUMINANCE32F',
    gl.GL_R32F                : 'GL_R32F',

    gl.GL_ALPHA8              : 'GL_ALPHA8',

    gl.GL_R8                  : 'GL_R8',
    gl.GL_R16                 : 'GL_R16',

    gl.GL_RGB8                : 'GL_RGB8',
    gl.GL_RGB16               : 'GL_RGB16',
    gl.GL_RGB32F              : 'GL_RGB32F',

    gl.GL_RGBA8               : 'GL_RGBA8',
    gl.GL_RGBA16              : 'GL_RGBA16',
    gl.GL_RGBA32F             : 'GL_RGBA32F',
}


def _makeInstance(dtype):
    """Used by :func:`oneChannelFormat` and :func:`getTextureType`.  If a
    ``numpy.dtype`` class is given, converts it into an instance.
    """
    if inspect.isclass(dtype):
        dtype = np.zeros([0], dtype=dtype).dtype
    return dtype


def numTextureDims(shape):
    """Given a 3D image shape, returns the number of dimensions needd
    to store the image as a texture - either ``2`` or ``3``.

    :arg shape: 3D image shape
    :returns:   2 if a 2D texture can be used to store the image data,
                3 otherwise.
    """
    max3d = gl.glGetInteger(gl.GL_MAX_3D_TEXTURE_SIZE)
    max2d = gl.glGetInteger(gl.GL_MAX_TEXTURE_SIZE)

    def checklim(shape, lim):
        if any([d > lim for d in shape]):
            raise RuntimeError(
                'Cannot create an OpenGL texture for {} - it exceeds '
                'the hardware limits on this platform (2D: {}, 3D: {}'
                .format(shape, max2d, max3d))

    shape = [d for d in shape[:3] if d > 1]

    # force scalar/vector shape to be 2D
    if   len(shape) == 0: shape = [1, 1]
    elif len(shape) == 1: shape = [shape[0], 1]

    if len(shape) == 3: checklim(shape, max3d)
    else:               checklim(shape, max2d)

    return len(shape)


@memoize.memoize
def canUseFloatTextures(nvals=1):
    """Returns ``True`` if this GL environment supports floating
    point textures, ``False`` otherwise. The test is based on the
    availability of the ``ARB_texture_float`` extension.

    :arg nvals: Number of values per voxel

    :returns: A tuple containing:

                - ``True`` if floating point textures are supported,
                  ``False`` otherwise

                - The base texture format to use (``None`` if floating
                  point textures are not supported)

                - The internal texture format to use  (``None`` if
                  floating point textures are not supported)
    """

    # We need the texture_float extension. The
    # texture_rg extension just provides some
    # nicer/more modern data types, but is not
    # necessary.
    floatSupported = glexts.hasExtension('GL_ARB_texture_float')
    rgSupported    = glexts.hasExtension('GL_ARB_texture_rg')

    if not floatSupported:
        return False, None, None

    if rgSupported:
        if   nvals == 1: baseFmt = gl.GL_RED
        elif nvals == 3: baseFmt = gl.GL_RGB
        elif nvals == 4: baseFmt = gl.GL_RGBA

        if   nvals == 1: intFmt  = gl.GL_R32F
        elif nvals == 3: intFmt  = gl.GL_RGB32F
        elif nvals == 4: intFmt  = gl.GL_RGBA32F

        return True, baseFmt, intFmt

    else:

        if   nvals == 1: baseFmt = gl.GL_LUMINANCE
        elif nvals == 3: baseFmt = gl.GL_RGB
        elif nvals == 4: baseFmt = gl.GL_RGBA

        if   nvals == 1: intFmt  = arbtf.GL_LUMINANCE32F_ARB
        elif nvals == 3: intFmt  = gl.GL_RGB32F
        elif nvals == 4: intFmt  = gl.GL_RGBA32F

        return True, baseFmt, intFmt


@memoize.memoize
def oneChannelFormat(dtype):
    """Determines suitable one-channel base and internal texture formats
    to use for the given ``numpy`` data type.

    :return: A tuple containing:

               - the base texture format to use
               - the internal texture format to use

    .. note:: This is used by the :func:`getTextureType` function. The
              returned formats may be ignored for floating point data
              types, depending on whether floating point textures are
              supported - see :func:`canUseFloatTextures`.
    """

    # Make sure we have a numpy dtype
    # *instance*, and not a class.
    dtype = _makeInstance(dtype)

    floatSupported = glexts.hasExtension('GL_ARB_texture_float')
    rgSupported    = glexts.hasExtension('GL_ARB_texture_rg')
    nbits          = dtype.itemsize * 8

    if rgSupported:
        if nbits == 8: return gl.GL_RED, gl.GL_R8
        else:          return gl.GL_RED, gl.GL_R16

    # GL_RED does not exist in old OpenGLs -
    # we have to use luminance instead.
    else:
        if nbits == 8:
            return gl.GL_LUMINANCE, gl.GL_LUMINANCE8

        # But GL_LUMINANCE is deprecated in GL 3.x,
        # and some more recent GL drivers seem to
        # have trouble displaying GL_LUMINANCE16
        # textures - displayting them at what seems
        # to be a down-sampled (e.g. using a 4 bit
        # storage format) version of the data. So we
        # store the data as floating point if we can.
        elif floatSupported:
            return gl.GL_LUMINANCE, arbtf.GL_LUMINANCE16F_ARB

        # Hopefully float textures are supported
        # on recent GL drivers which don't support
        # LUMINANCE_16
        else:
            return gl.GL_LUMINANCE, gl.GL_LUMINANCE16


@memoize.memoize
def getTextureType(normalise, dtype, nvals):
    """Figures out the GL data type, and the base/internal texture
    formats in which the specified data should be stored.


    This function just figures out the data types which should be used to
    store the data as a texture. Any necessary data conversion/transformation
    is performed by the :func:`prepareData` method.


    The data can be stored as a GL texture as-is if:

      - it is of type ``uint8`` or ``uint16``

      - it is of type ``float32``, *and* this GL environment has support
        for floating point textures. Support for floating point textures
        is determined by the :func:`canUseFlotaTextures` function.


    In all other cases, the data needs to be converted to a supported data
    type, and potentially normalised, before it can be used as texture
    data.  If floating point textures are available, the data is converted
    to float32. Otherwise, the data is converted to ``uint16``, and
    normalised to take the full ``uint16`` data range (``0-65535``), and a
    transformation matrix is saved, allowing transformation of this
    normalised data back to its original data range.


    .. note:: OpenGL does different things to texture data depending on
              its type: unsigned integer types are normalised from ``[0,
              INT_MAX]`` to ``[0, 1]``.

              Floating point texture data types are, by default, *clamped*
              (not normalised), to the range ``[0, 1]``! We can overcome by
              this by using a true floating point texture, which is
              accomplished by using one of the data types provided by the
              ``ARB_texture_float`` extension. If this extension is not
              available, we have no choice but to normalise the data.


    :arg normalise: Whether the data is to be normalised or not

    :arg dtype:     The original data type (e.g. ``np.uint8``)

    :arg nvals:     Number of values per voxel. Must be either ``1``,
                    ``3``, or ``4``.

    :returns:       A tuple containing:

                     - The raw type of the texture data
                       (e.g. ``GL_UNSIGNED_SHORT``)

                     - The texture format (e.g. ``GL_RGB``, ``GL_LUMINANCE``,
                       etc).

                     - The internal texture format used by OpenGL for storage
                       (e.g. ``GL_RGB16``, ``GL_LUMINANCE8``, etc).
    """

    dtype = _makeInstance(dtype)

    floatTextures, fBaseFmt, fIntFmt = canUseFloatTextures(nvals)
    ocBaseFmt, ocIntFmt              = oneChannelFormat(dtype)
    isFloat                          = issubclass(dtype.type, np.floating)

    # Signed data types are a pain in the arse.
    # We have to store them as unsigned, and
    # apply an offset.

    # Note: Throughout this function, it is assumed
    #       that if the data type is not supported,
    #       then the normalise flag will have been
    #       set to True.

    # Data type
    if   normalise:          texDtype = gl.GL_UNSIGNED_SHORT
    elif dtype == np.uint8:  texDtype = gl.GL_UNSIGNED_BYTE
    elif dtype == np.int8:   texDtype = gl.GL_UNSIGNED_BYTE
    elif dtype == np.uint16: texDtype = gl.GL_UNSIGNED_SHORT
    elif dtype == np.int16:  texDtype = gl.GL_UNSIGNED_SHORT
    elif floatTextures:      texDtype = gl.GL_FLOAT

    # Base texture format
    if floatTextures and isFloat: baseFmt = fBaseFmt
    elif nvals == 1:              baseFmt = ocBaseFmt
    elif nvals == 3:              baseFmt = gl.GL_RGB
    elif nvals == 4:              baseFmt = gl.GL_RGBA

    # Internal texture format
    if nvals == 1:
        if   normalise:          intFmt = ocIntFmt
        elif dtype == np.uint8:  intFmt = ocIntFmt
        elif dtype == np.int8:   intFmt = ocIntFmt
        elif dtype == np.uint16: intFmt = ocIntFmt
        elif dtype == np.int16:  intFmt = ocIntFmt
        elif floatTextures:      intFmt = fIntFmt

    elif nvals == 3:
        if   normalise:          intFmt = gl.GL_RGB16
        elif dtype == np.uint8:  intFmt = gl.GL_RGB8
        elif dtype == np.int8:   intFmt = gl.GL_RGB8
        elif dtype == np.uint16: intFmt = gl.GL_RGB16
        elif dtype == np.int16:  intFmt = gl.GL_RGB16
        elif floatTextures:      intFmt = fIntFmt

    elif nvals == 4:
        if   normalise:          intFmt = gl.GL_RGBA16
        elif dtype == np.uint8:  intFmt = gl.GL_RGBA8
        elif dtype == np.int8:   intFmt = gl.GL_RGBA8
        elif dtype == np.uint16: intFmt = gl.GL_RGBA16
        elif dtype == np.int16:  intFmt = gl.GL_RGBA16
        elif floatTextures:      intFmt = fIntFmt

    return texDtype, baseFmt, intFmt


def prepareData(data,
                prefilter=None,
                prefilterRange=None,
                resolution=None,
                scales=None,
                normalise=None,
                normaliseRange=None):
    """This function prepares and returns the given ``data``, ready to be
    used as GL texture data.

    This process potentially involves:

      - Resampling to a different resolution (see the
        :func:`.routines.subsample` function).

      - Pre-filtering (see the ``prefilter`` parameter to
        :meth:`__init__`).

      - Normalising (if the ``normalise`` parameter to :meth:`__init__`
        was ``True``, or if the data type cannot be used as-is).

      - Casting to a different data type (if the data type cannot be used
        as-is).

    :returns: A tuple containing:

                - A ``numpy`` array containing the image data, ready to be
                   copied to the GPU.

                - An affine transformation matrix which encodes an offset
                  and a scale, which may be used to transform the texture
                  data from the range ``[0.0, 1.0]`` to its raw data
                  range.

                - Inverse of ``voxValXform``.
    """

    dtype         = data.dtype
    floatTextures = canUseFloatTextures()

    if normalise: dmin, dmax = normaliseRange
    else:         dmin, dmax = 0, 0

    if normalise                  and \
       prefilter      is not None and \
       prefilterRange is not None:
        dmin, dmax = prefilterRange(dmin, dmax)

    # Offsets/scales which can be used to transform from
    # the texture data (which may be offset or normalised)
    # back to the original voxel data
    if   normalise:          offset =  dmin
    elif dtype == np.uint8:  offset =  0
    elif dtype == np.int8:   offset = -128
    elif dtype == np.uint16: offset =  0
    elif dtype == np.int16:  offset = -32768
    elif floatTextures:      offset = 0

    if   normalise:          scale = dmax - dmin
    elif dtype == np.uint8:  scale = 255
    elif dtype == np.int8:   scale = 255
    elif dtype == np.uint16: scale = 65535
    elif dtype == np.int16:  scale = 65535
    elif floatTextures:      scale = 1

    # If the data range is 0 (min == max)
    # we just set an identity xform
    if scale == 0:
        voxValXform    = np.eye(4)
        invVoxValXform = np.eye(4)

    # Otherwise we save a transformation
    # from the texture values back to the
    # original data range. Note that if
    # storing floating point data, this
    # will be an identity transform.
    else:
        invScale       = 1.0 / scale
        voxValXform    = affine.scaleOffsetXform(scale, offset)
        invVoxValXform = affine.scaleOffsetXform(
            invScale,
            -offset * invScale)

    if resolution is not None:
        data = glroutines.subsample(data, resolution, pixdim=scales)[0]

    if prefilter is not None:
        data = prefilter(data)

    # TODO if FLOAT_TEXTURES, you should
    #      save normalised values as float32
    if normalise:

        log.debug('Normalising to range {} - {}'.format(dmin, dmax))

        if dmax != dmin:
            data = np.clip((data - dmin) / float(dmax - dmin), 0, 1)

        data = np.round(data * 65535)
        data = np.array(data, dtype=np.uint16)

    elif dtype == np.uint8:  pass
    elif dtype == np.int8:   data = np.array(data + 128,   dtype=np.uint8)
    elif dtype == np.uint16: pass
    elif dtype == np.int16:  data = np.array(data + 32768, dtype=np.uint16)
    elif floatTextures and data.dtype != np.float32:
        data = np.array(data, dtype=np.float32)

    return data, voxValXform, invVoxValXform
